# https://etherscan.io/address/0x03403154afc09Ce8e44C3B185C82C6aD5f86b9ab
YVAULT_ABI = [
    {'inputs': [{'internalType': 'address', 'name': '_token', 'type': 'address'},
                {'internalType': 'address', 'name': '_controller', 'type': 'address'}],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'constructor'},
    {'anonymous': False,
     'inputs': [{'indexed': True,
                 'internalType': 'address',
                 'name': 'owner',
                 'type': 'address'},
                {'indexed': True,
                 'internalType': 'address',
                 'name': 'spender',
                 'type': 'address'},
                {'indexed': False,
                 'internalType': 'uint256',
                 'name': 'value',
                 'type': 'uint256'}],
     'name': 'Approval',
     'type': 'event'},
    {'anonymous': False,
     'inputs': [{'indexed': True,
                 'internalType': 'address',
                 'name': 'from',
                 'type': 'address'},
                {'indexed': True,
                 'internalType': 'address',
                 'name': 'to',
                 'type': 'address'},
                {'indexed': False,
                 'internalType': 'uint256',
                 'name': 'value',
                 'type': 'uint256'}],
     'name': 'Transfer',
     'type': 'event'},
    {'constant': True,
     'inputs': [{'internalType': 'address', 'name': 'owner', 'type': 'address'},
                {'internalType': 'address', 'name': 'spender', 'type': 'address'}],
     'name': 'allowance',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'payable': False,
     'stateMutability': 'view',
     'type': 'function'},
    {'constant': False,
     'inputs': [{'internalType': 'address', 'name': 'spender', 'type': 'address'},
                {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'}],
     'name': 'approve',
     'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'constant': True,
     'inputs': [],
     'name': 'available',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'payable': False,
     'stateMutability': 'view',
     'type': 'function'},
    {'constant': True,
     'inputs': [],
     'name': 'balance',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'payable': False,
     'stateMutability': 'view',
     'type': 'function'},
    {'constant': True,
     'inputs': [{'internalType': 'address',
                 'name': 'account',
                 'type': 'address'}],
     'name': 'balanceOf',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'payable': False,
     'stateMutability': 'view',
     'type': 'function'},
    {'constant': True,
     'inputs': [],
     'name': 'controller',
     'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
     'payable': False,
     'stateMutability': 'view',
     'type': 'function'},
    {'constant': True,
     'inputs': [],
     'name': 'decimals',
     'outputs': [{'internalType': 'uint8', 'name': '', 'type': 'uint8'}],
     'payable': False,
     'stateMutability': 'view',
     'type': 'function'},
    {'constant': False,
     'inputs': [{'internalType': 'address', 'name': 'spender', 'type': 'address'},
                {'internalType': 'uint256', 'name': 'subtractedValue', 'type': 'uint256'}],
     'name': 'decreaseAllowance',
     'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'constant': False,
     'inputs': [{'internalType': 'uint256',
                 'name': '_amount',
                 'type': 'uint256'}],
     'name': 'deposit',
     'outputs': [],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'constant': False,
     'inputs': [],
     'name': 'depositAll',
     'outputs': [],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'constant': False,
     'inputs': [],
     'name': 'earn',
     'outputs': [],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'constant': True,
     'inputs': [],
     'name': 'getPricePerFullShare',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'payable': False,
     'stateMutability': 'view',
     'type': 'function'},
    {'constant': True,
     'inputs': [],
     'name': 'governance',
     'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
     'payable': False,
     'stateMutability': 'view',
     'type': 'function'},
    {'constant': False,
     'inputs': [{'internalType': 'address', 'name': 'spender', 'type': 'address'},
                {'internalType': 'uint256', 'name': 'addedValue', 'type': 'uint256'}],
     'name': 'increaseAllowance',
     'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'constant': True,
     'inputs': [],
     'name': 'max',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'payable': False,
     'stateMutability': 'view',
     'type': 'function'},
    {'constant': True,
     'inputs': [],
     'name': 'min',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'payable': False,
     'stateMutability': 'view',
     'type': 'function'},
    {'constant': True,
     'inputs': [],
     'name': 'name',
     'outputs': [{'internalType': 'string', 'name': '', 'type': 'string'}],
     'payable': False,
     'stateMutability': 'view',
     'type': 'function'},
    {'constant': False,
     'inputs': [{'internalType': 'address',
                 'name': '_controller',
                 'type': 'address'}],
     'name': 'setController',
     'outputs': [],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'constant': False,
     'inputs': [{'internalType': 'address',
                 'name': '_governance',
                 'type': 'address'}],
     'name': 'setGovernance',
     'outputs': [],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'constant': False,
     'inputs': [{'internalType': 'uint256', 'name': '_min', 'type': 'uint256'}],
     'name': 'setMin',
     'outputs': [],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'constant': False,
     'inputs': [{'internalType': 'string', 'name': '_name', 'type': 'string'}],
     'name': 'setName',
     'outputs': [],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'constant': False,
     'inputs': [{'internalType': 'string', 'name': '_symbol', 'type': 'string'}],
     'name': 'setSymbol',
     'outputs': [],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'constant': True,
     'inputs': [],
     'name': 'symbol',
     'outputs': [{'internalType': 'string', 'name': '', 'type': 'string'}],
     'payable': False,
     'stateMutability': 'view',
     'type': 'function'},
    {'constant': True,
     'inputs': [],
     'name': 'token',
     'outputs': [{'internalType': 'contract IERC20',
                  'name': '',
                  'type': 'address'}],
     'payable': False,
     'stateMutability': 'view',
     'type': 'function'},
    {'constant': True,
     'inputs': [],
     'name': 'totalSupply',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'payable': False,
     'stateMutability': 'view',
     'type': 'function'},
    {'constant': False,
     'inputs': [{'internalType': 'address',
                 'name': 'recipient',
                 'type': 'address'},
                {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'}],
     'name': 'transfer',
     'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'constant': False,
     'inputs': [{'internalType': 'address', 'name': 'sender', 'type': 'address'},
                {'internalType': 'address', 'name': 'recipient', 'type': 'address'},
                {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'}],
     'name': 'transferFrom',
     'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'constant': False,
     'inputs': [{'internalType': 'uint256',
                 'name': '_shares',
                 'type': 'uint256'}],
     'name': 'withdraw',
     'outputs': [],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'constant': False,
     'inputs': [],
     'name': 'withdrawAll',
     'outputs': [],
     'payable': False,
     'stateMutability': 'nonpayable',
     'type': 'function'}
]

YTOKEN_ABI = [
    {'name': 'Transfer',
     'inputs': [{'name': 'sender', 'type': 'address', 'indexed': True},
                {'name': 'receiver', 'type': 'address', 'indexed': True},
                {'name': 'value', 'type': 'uint256', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'Approval',
     'inputs': [{'name': 'owner', 'type': 'address', 'indexed': True},
                {'name': 'spender', 'type': 'address', 'indexed': True},
                {'name': 'value', 'type': 'uint256', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'StrategyAdded',
     'inputs': [{'name': 'strategy', 'type': 'address', 'indexed': True},
                {'name': 'debtRatio', 'type': 'uint256', 'indexed': False},
                {'name': 'minDebtPerHarvest', 'type': 'uint256', 'indexed': False},
                {'name': 'maxDebtPerHarvest', 'type': 'uint256', 'indexed': False},
                {'name': 'performanceFee', 'type': 'uint256', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'StrategyReported',
     'inputs': [{'name': 'strategy', 'type': 'address', 'indexed': True},
                {'name': 'gain', 'type': 'uint256', 'indexed': False},
                {'name': 'loss', 'type': 'uint256', 'indexed': False},
                {'name': 'debtPaid', 'type': 'uint256', 'indexed': False},
                {'name': 'totalGain', 'type': 'uint256', 'indexed': False},
                {'name': 'totalLoss', 'type': 'uint256', 'indexed': False},
                {'name': 'totalDebt', 'type': 'uint256', 'indexed': False},
                {'name': 'debtAdded', 'type': 'uint256', 'indexed': False},
                {'name': 'debtRatio', 'type': 'uint256', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'UpdateGovernance',
     'inputs': [{'name': 'governance', 'type': 'address', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'UpdateManagement',
     'inputs': [{'name': 'management', 'type': 'address', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'UpdateGuestList',
     'inputs': [{'name': 'guestList', 'type': 'address', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'UpdateRewards',
     'inputs': [{'name': 'rewards', 'type': 'address', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'UpdateDepositLimit',
     'inputs': [{'name': 'depositLimit', 'type': 'uint256', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'UpdatePerformanceFee',
     'inputs': [{'name': 'performanceFee', 'type': 'uint256', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'UpdateManagementFee',
     'inputs': [{'name': 'managementFee', 'type': 'uint256', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'UpdateGuardian',
     'inputs': [{'name': 'guardian', 'type': 'address', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'EmergencyShutdown',
     'inputs': [{'name': 'active', 'type': 'bool', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'UpdateWithdrawalQueue',
     'inputs': [{'name': 'queue', 'type': 'address[20]', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'StrategyUpdateDebtRatio',
     'inputs': [{'name': 'strategy', 'type': 'address', 'indexed': True},
                {'name': 'debtRatio', 'type': 'uint256', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'StrategyUpdateMinDebtPerHarvest',
     'inputs': [{'name': 'strategy', 'type': 'address', 'indexed': True},
                {'name': 'minDebtPerHarvest', 'type': 'uint256', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'StrategyUpdateMaxDebtPerHarvest',
     'inputs': [{'name': 'strategy', 'type': 'address', 'indexed': True},
                {'name': 'maxDebtPerHarvest', 'type': 'uint256', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'StrategyUpdatePerformanceFee',
     'inputs': [{'name': 'strategy', 'type': 'address', 'indexed': True},
                {'name': 'performanceFee', 'type': 'uint256', 'indexed': False}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'StrategyMigrated',
     'inputs': [{'name': 'oldVersion', 'type': 'address', 'indexed': True},
                {'name': 'newVersion', 'type': 'address', 'indexed': True}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'StrategyRevoked',
     'inputs': [{'name': 'strategy', 'type': 'address', 'indexed': True}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'StrategyRemovedFromQueue',
     'inputs': [{'name': 'strategy', 'type': 'address', 'indexed': True}],
     'anonymous': False,
     'type': 'event'},
    {'name': 'StrategyAddedToQueue',
     'inputs': [{'name': 'strategy', 'type': 'address', 'indexed': True}],
     'anonymous': False,
     'type': 'event'},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'initialize',
     'inputs': [{'name': 'token', 'type': 'address'},
                {'name': 'governance', 'type': 'address'},
                {'name': 'rewards', 'type': 'address'},
                {'name': 'nameOverride', 'type': 'string'},
                {'name': 'symbolOverride', 'type': 'string'}],
     'outputs': []},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'initialize',
     'inputs': [{'name': 'token', 'type': 'address'},
                {'name': 'governance', 'type': 'address'},
                {'name': 'rewards', 'type': 'address'},
                {'name': 'nameOverride', 'type': 'string'},
                {'name': 'symbolOverride', 'type': 'string'},
                {'name': 'guardian', 'type': 'address'}],
     'outputs': []},
    {'stateMutability': 'pure',
     'type': 'function',
     'name': 'apiVersion',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'string'}],
     'gas': 4546},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'setName',
     'inputs': [{'name': 'name', 'type': 'string'}],
     'outputs': [],
     'gas': 107044},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'setSymbol',
     'inputs': [{'name': 'symbol', 'type': 'string'}],
     'outputs': [],
     'gas': 71894},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'setGovernance',
     'inputs': [{'name': 'governance', 'type': 'address'}],
     'outputs': [],
     'gas': 36365},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'acceptGovernance',
     'inputs': [],
     'outputs': [],
     'gas': 37637},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'setManagement',
     'inputs': [{'name': 'management', 'type': 'address'}],
     'outputs': [],
     'gas': 37775},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'setGuestList',
     'inputs': [{'name': 'guestList', 'type': 'address'}],
     'outputs': [],
     'gas': 37805},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'setRewards',
     'inputs': [{'name': 'rewards', 'type': 'address'}],
     'outputs': [],
     'gas': 37835},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'setLockedProfitDegration',
     'inputs': [{'name': 'degration', 'type': 'uint256'}],
     'outputs': [],
     'gas': 36519},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'setDepositLimit',
     'inputs': [{'name': 'limit', 'type': 'uint256'}],
     'outputs': [],
     'gas': 37795},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'setPerformanceFee',
     'inputs': [{'name': 'fee', 'type': 'uint256'}],
     'outputs': [],
     'gas': 37929},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'setManagementFee',
     'inputs': [{'name': 'fee', 'type': 'uint256'}],
     'outputs': [],
     'gas': 37959},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'setGuardian',
     'inputs': [{'name': 'guardian', 'type': 'address'}],
     'outputs': [],
     'gas': 39203},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'setEmergencyShutdown',
     'inputs': [{'name': 'active', 'type': 'bool'}],
     'outputs': [],
     'gas': 39274},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'setWithdrawalQueue',
     'inputs': [{'name': 'queue', 'type': 'address[20]'}],
     'outputs': [],
     'gas': 763950},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'transfer',
     'inputs': [{'name': 'receiver', 'type': 'address'},
                {'name': 'amount', 'type': 'uint256'}],
     'outputs': [{'name': '', 'type': 'bool'}],
     'gas': 76768},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'transferFrom',
     'inputs': [{'name': 'sender', 'type': 'address'},
                {'name': 'receiver', 'type': 'address'},
                {'name': 'amount', 'type': 'uint256'}],
     'outputs': [{'name': '', 'type': 'bool'}],
     'gas': 116531},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'approve',
     'inputs': [{'name': 'spender', 'type': 'address'},
                {'name': 'amount', 'type': 'uint256'}],
     'outputs': [{'name': '', 'type': 'bool'}],
     'gas': 38271},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'increaseAllowance',
     'inputs': [{'name': 'spender', 'type': 'address'},
                {'name': 'amount', 'type': 'uint256'}],
     'outputs': [{'name': '', 'type': 'bool'}],
     'gas': 40312},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'decreaseAllowance',
     'inputs': [{'name': 'spender', 'type': 'address'},
                {'name': 'amount', 'type': 'uint256'}],
     'outputs': [{'name': '', 'type': 'bool'}],
     'gas': 40336},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'permit',
     'inputs': [{'name': 'owner', 'type': 'address'},
                {'name': 'spender', 'type': 'address'},
                {'name': 'amount', 'type': 'uint256'},
                {'name': 'expiry', 'type': 'uint256'},
                {'name': 'signature', 'type': 'bytes'}],
     'outputs': [{'name': '', 'type': 'bool'}],
     'gas': 81264},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'totalAssets',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 4098},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'deposit',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}]},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'deposit',
     'inputs': [{'name': '_amount', 'type': 'uint256'}],
     'outputs': [{'name': '', 'type': 'uint256'}]},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'deposit',
     'inputs': [{'name': '_amount', 'type': 'uint256'},
                {'name': 'recipient', 'type': 'address'}],
     'outputs': [{'name': '', 'type': 'uint256'}]},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'maxAvailableShares',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 383839},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'withdraw',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}]},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'withdraw',
     'inputs': [{'name': 'maxShares', 'type': 'uint256'}],
     'outputs': [{'name': '', 'type': 'uint256'}]},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'withdraw',
     'inputs': [{'name': 'maxShares', 'type': 'uint256'},
                {'name': 'recipient', 'type': 'address'}],
     'outputs': [{'name': '', 'type': 'uint256'}]},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'withdraw',
     'inputs': [{'name': 'maxShares', 'type': 'uint256'},
                {'name': 'recipient', 'type': 'address'},
                {'name': 'maxLoss', 'type': 'uint256'}],
     'outputs': [{'name': '', 'type': 'uint256'}]},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'pricePerShare',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 18195},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'addStrategy',
     'inputs': [{'name': 'strategy', 'type': 'address'},
                {'name': 'debtRatio', 'type': 'uint256'},
                {'name': 'minDebtPerHarvest', 'type': 'uint256'},
                {'name': 'maxDebtPerHarvest', 'type': 'uint256'},
                {'name': 'performanceFee', 'type': 'uint256'}],
     'outputs': [],
     'gas': 1485796},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'updateStrategyDebtRatio',
     'inputs': [{'name': 'strategy', 'type': 'address'},
                {'name': 'debtRatio', 'type': 'uint256'}],
     'outputs': [],
     'gas': 115193},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'updateStrategyMinDebtPerHarvest',
     'inputs': [{'name': 'strategy', 'type': 'address'},
                {'name': 'minDebtPerHarvest', 'type': 'uint256'}],
     'outputs': [],
     'gas': 42441},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'updateStrategyMaxDebtPerHarvest',
     'inputs': [{'name': 'strategy', 'type': 'address'},
                {'name': 'maxDebtPerHarvest', 'type': 'uint256'}],
     'outputs': [],
     'gas': 42471},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'updateStrategyPerformanceFee',
     'inputs': [{'name': 'strategy', 'type': 'address'},
                {'name': 'performanceFee', 'type': 'uint256'}],
     'outputs': [],
     'gas': 41251},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'migrateStrategy',
     'inputs': [{'name': 'oldVersion', 'type': 'address'},
                {'name': 'newVersion', 'type': 'address'}],
     'outputs': [],
     'gas': 1141468},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'revokeStrategy',
     'inputs': [],
     'outputs': []},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'revokeStrategy',
     'inputs': [{'name': 'strategy', 'type': 'address'}],
     'outputs': []},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'addStrategyToQueue',
     'inputs': [{'name': 'strategy', 'type': 'address'}],
     'outputs': [],
     'gas': 1199804},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'removeStrategyFromQueue',
     'inputs': [{'name': 'strategy', 'type': 'address'}],
     'outputs': [],
     'gas': 23088703},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'debtOutstanding',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}]},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'debtOutstanding',
     'inputs': [{'name': 'strategy', 'type': 'address'}],
     'outputs': [{'name': '', 'type': 'uint256'}]},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'creditAvailable',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}]},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'creditAvailable',
     'inputs': [{'name': 'strategy', 'type': 'address'}],
     'outputs': [{'name': '', 'type': 'uint256'}]},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'availableDepositLimit',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 9551},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'expectedReturn',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}]},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'expectedReturn',
     'inputs': [{'name': 'strategy', 'type': 'address'}],
     'outputs': [{'name': '', 'type': 'uint256'}]},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'report',
     'inputs': [{'name': 'gain', 'type': 'uint256'},
                {'name': 'loss', 'type': 'uint256'},
                {'name': '_debtPayment', 'type': 'uint256'}],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 1015170},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'sweep',
     'inputs': [{'name': 'token', 'type': 'address'}],
     'outputs': []},
    {'stateMutability': 'nonpayable',
     'type': 'function',
     'name': 'sweep',
     'inputs': [{'name': 'token', 'type': 'address'},
                {'name': 'amount', 'type': 'uint256'}],
     'outputs': []},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'name',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'string'}],
     'gas': 8750},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'symbol',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'string'}],
     'gas': 7803},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'decimals',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 2408},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'precisionFactor',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 2438},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'balanceOf',
     'inputs': [{'name': 'arg0', 'type': 'address'}],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 2683},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'allowance',
     'inputs': [{'name': 'arg0', 'type': 'address'},
                {'name': 'arg1', 'type': 'address'}],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 2928},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'totalSupply',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 2528},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'token',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'address'}],
     'gas': 2558},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'governance',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'address'}],
     'gas': 2588},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'management',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'address'}],
     'gas': 2618},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'guardian',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'address'}],
     'gas': 2648},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'guestList',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'address'}],
     'gas': 2678},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'strategies',
     'inputs': [{'name': 'arg0', 'type': 'address'}],
     'outputs': [{'name': 'performanceFee', 'type': 'uint256'},
                 {'name': 'activation', 'type': 'uint256'},
                 {'name': 'debtRatio', 'type': 'uint256'},
                 {'name': 'minDebtPerHarvest', 'type': 'uint256'},
                 {'name': 'maxDebtPerHarvest', 'type': 'uint256'},
                 {'name': 'lastReport', 'type': 'uint256'},
                 {'name': 'totalDebt', 'type': 'uint256'},
                 {'name': 'totalGain', 'type': 'uint256'},
                 {'name': 'totalLoss', 'type': 'uint256'}],
     'gas': 11031},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'withdrawalQueue',
     'inputs': [{'name': 'arg0', 'type': 'uint256'}],
     'outputs': [{'name': '', 'type': 'address'}],
     'gas': 2847},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'emergencyShutdown',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'bool'}],
     'gas': 2768},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'depositLimit',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 2798},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'debtRatio',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 2828},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'totalDebt',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 2858},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'lastReport',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 2888},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'activation',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 2918},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'lockedProfit',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 2948},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'lockedProfitDegration',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 2978},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'rewards',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'address'}],
     'gas': 3008},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'managementFee',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 3038},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'performanceFee',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 3068},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'nonces',
     'inputs': [{'name': 'arg0', 'type': 'address'}],
     'outputs': [{'name': '', 'type': 'uint256'}],
     'gas': 3313},
    {'stateMutability': 'view',
     'type': 'function',
     'name': 'DOMAIN_SEPARATOR',
     'inputs': [],
     'outputs': [{'name': '', 'type': 'bytes32'}],
     'gas': 3128}
]



import os
import random
from logging import getLogger

from ethereum.transactions import secpk1n
from web3 import Web3

from gnosis.eth.tests.utils import send_tx

from ..safe_creation_tx import SafeCreationTx

logger = getLogger(__name__)


def generate_salt_nonce() -> int:
    return random.getrandbits(256) - 1


def generate_valid_s() -> int:
    while True:
        s = int(os.urandom(30).hex(), 16)
        if s <= (secpk1n // 2):
            return s


def deploy_safe(w3: Web3, safe_creation_tx: SafeCreationTx, funder: str, initial_funding_wei: int = 0,
                funder_account=None) -> str:
    if funder_account:
        send_tx(w3, {
            'to': safe_creation_tx.deployer_address,
            'value': safe_creation_tx.payment,
        }, funder_account)

        send_tx(w3, {
            'to': safe_creation_tx.safe_address,
            'value': safe_creation_tx.payment + initial_funding_wei,
        }, funder_account)
    else:
        w3.eth.wait_for_transaction_receipt(
            w3.eth.send_transaction({
                'from': funder,
                'to': safe_creation_tx.deployer_address,
                'value': safe_creation_tx.payment
            })
        )

        w3.eth.wait_for_transaction_receipt(
            w3.eth.send_transaction({
                'from': funder,
                'to': safe_creation_tx.safe_address,
                'value': safe_creation_tx.payment + initial_funding_wei
            })
        )

    tx_hash = w3.eth.send_raw_transaction(bytes(safe_creation_tx.tx_raw))
    tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
    assert tx_receipt.contractAddress == safe_creation_tx.safe_address
    assert tx_receipt.status

    return safe_creation_tx.safe_address



import json
import time
from typing import Any, Dict, Optional
from urllib.parse import urljoin

import requests

from .. import EthereumNetwork
from .contract_metadata import ContractMetadata


class EtherscanClientException(Exception):
    pass


class EtherscanClientConfigurationProblem(Exception):
    pass


class EtherscanRateLimitError(EtherscanClientException):
    pass


class EtherscanClient:
    NETWORK_WITH_URL = {
        EthereumNetwork.MAINNET: 'https://etherscan.io',
        EthereumNetwork.RINKEBY: 'https://rinkeby.etherscan.io',
        EthereumNetwork.ROPSTEN: 'https://ropsten.etherscan.io',
        EthereumNetwork.GOERLI: 'https://goerli.etherscan.io',
        EthereumNetwork.KOVAN: 'https://kovan.etherscan.io',
        EthereumNetwork.BINANCE: 'https://bscscan.com',
        EthereumNetwork.MATIC: 'https://polygonscan.com',
        EthereumNetwork.OPTIMISTIC: 'https://optimistic.etherscan.io',
    }

    NETWORK_WITH_API_URL = {
        EthereumNetwork.MAINNET: 'https://api.etherscan.io',
        EthereumNetwork.RINKEBY: 'https://api-rinkeby.etherscan.io',
        EthereumNetwork.ROPSTEN: 'https://api-ropsten.etherscan.io',
        EthereumNetwork.GOERLI: 'https://api-goerli.etherscan.io/',
        EthereumNetwork.KOVAN: 'https://api-kovan.etherscan.io/',
        EthereumNetwork.BINANCE: 'https://api.bscscan.com',
        EthereumNetwork.MATIC: 'https://api.polygonscan.com',
        EthereumNetwork.OPTIMISTIC: 'https://api-optimistic.etherscan.io',
    }
    HTTP_HEADERS = {
        'User-Agent': 'curl/7.77.0',
    }

    def __init__(self, network: EthereumNetwork, api_key: Optional[str] = None):
        self.network = network
        self.api_key = api_key
        self.base_url = self.NETWORK_WITH_URL.get(network)
        self.base_api_url = self.NETWORK_WITH_API_URL.get(network)
        if self.base_api_url is None:
            raise EtherscanClientConfigurationProblem(f'Network {network.name} - {network.value} not supported')
        self.http_session = requests.Session()
        self.http_session.headers = self.HTTP_HEADERS

    def build_url(self, path: str):
        url = urljoin(self.base_api_url, path)
        if self.api_key:
            url += f'&apikey={self.api_key}'
        return url

    def _do_request(self, url: str) -> Optional[Dict[str, Any]]:
        response = self.http_session.get(url)

        if response.ok:
            response_json = response.json()
            result = response_json['result']
            if 'Max rate limit reached' in result:
                # Max rate limit reached, please use API Key for higher rate limit
                raise EtherscanRateLimitError
            if response_json['status'] == '1':
                return result

    def _retry_request(self, url: str, retry: bool = True) -> Optional[Dict[str, Any]]:
        for _ in range(3):
            try:
                return self._do_request(url)
            except EtherscanRateLimitError as exc:
                if not retry:
                    raise exc
                else:
                    time.sleep(5)

    def get_contract_metadata(self, contract_address: str, retry: bool = True) -> Optional[ContractMetadata]:
        contract_source_code = self.get_contract_source_code(contract_address, retry=retry)
        if contract_source_code:
            contract_name = contract_source_code['ContractName']
            contract_abi = contract_source_code['ABI']
            if contract_abi:
                return ContractMetadata(contract_name, contract_abi, False)

    def get_contract_source_code(self, contract_address: str, retry: bool = True):
        """
        Get source code for a contract. Source code query also returns:

            - ContractName: "",
            - CompilerVersion: "",
            - OptimizationUsed: "",
            - Runs: "",
            - ConstructorArguments: ""
            - EVMVersion: "Default",
            - Library: "",
            - LicenseType: "",
            - Proxy: "0",
            - Implementation: "",
            - SwarmSource: ""

        :param contract_address:
        :param retry: if ``True``, try again if there's Rate Limit Error
        :return:
        """
        url = self.build_url(f'api?module=contract&action=getsourcecode&address={contract_address}')
        result = self._retry_request(url, retry=retry)  # Returns a list
        if result:
            result = result[0]
            abi_str = result['ABI']
            result['ABI'] = json.loads(abi_str) if abi_str.startswith('[') else None
            return result

    def get_contract_abi(self, contract_address: str, retry: bool = True):
        url = self.build_url(f'api?module=contract&action=getabi&address={contract_address}')
        result = self._retry_request(url, retry=retry)
        if result:
            return json.loads(result)



from django.core.exceptions import ValidationError

from web3 import Web3


def validate_checksumed_address(address):
    if not Web3.isChecksumAddress(address):
        raise ValidationError(
            '%(address)s has an invalid checksum',
            params={'address': address},
        )



from django.db import models
from django.utils.translation import gettext_lazy as _

import ethereum.utils
from hexbytes import HexBytes

from .validators import validate_checksumed_address

try:
    from django.db import DefaultConnectionProxy
    connection = DefaultConnectionProxy()
except ImportError:
    from django.db import connections
    connection = connections['default']


class EthereumAddressField(models.CharField):
    default_validators = [validate_checksumed_address]
    description = "Ethereum address"

    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = 42
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs['max_length']
        return name, path, args, kwargs

    def from_db_value(self, value, expression, connection):
        return self.to_python(value)

    def to_python(self, value):
        value = super().to_python(value)
        if value:
            return ethereum.utils.checksum_encode(value)
        else:
            return value

    def get_prep_value(self, value):
        value = super().get_prep_value(value)
        if value:
            return ethereum.utils.checksum_encode(value)
        else:
            return value


class Uint256Field(models.DecimalField):
    description = _("Ethereum uint256 number")
    """
    Field to store ethereum uint256 values. Uses Decimal db type without decimals to store
    in the database, but retrieve as `int` instead of `Decimal` (https://docs.python.org/3/library/decimal.html)
    """
    def __init__(self, *args, **kwargs):
        kwargs['max_digits'] = 79  # 2 ** 256 is 78 digits
        kwargs['decimal_places'] = 0
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs['max_digits']
        del kwargs['decimal_places']
        return name, path, args, kwargs

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return int(value)


class HexField(models.CharField):
    """
    Field to store hex values (without 0x). Returns hex with 0x prefix.

    On Database side a CharField is used.
    """
    description = "Stores a hex value into an CharField"

    def from_db_value(self, value, expression, connection):
        return self.to_python(value)

    def to_python(self, value):
        return value if value is None else HexBytes(value).hex()

    def get_prep_value(self, value):
        if value is None:
            return value
        elif isinstance(value, HexBytes):
            return value.hex()[2:]  # HexBytes.hex() retrieves hexadecimal with '0x', remove it
        elif isinstance(value, bytes):
            return value.hex()  # bytes.hex() retrieves hexadecimal without '0x'
        else:  # str
            return HexBytes(value).hex()[2:]

    def formfield(self, **kwargs):
        # We need max_lenght + 2 on forms because of `0x`
        defaults = {'max_length': self.max_length + 2}
        # TODO: Handle multiple backends with different feature flags.
        if self.null and not connection.features.interprets_empty_strings_as_nulls:
            defaults['empty_value'] = None
        defaults.update(kwargs)
        return super().formfield(**defaults)

    def clean(self, value, model_instance):
        value = self.to_python(value)
        self.validate(value, model_instance)
        # Validation didn't work because of `0x`
        self.run_validators(value[2:])
        return value


class Sha3HashField(HexField):
    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = 64
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs['max_length']
        return name, path, args, kwargs



from django.test import TestCase

from ... import EthereumNetwork
from ...clients import BlockscoutClient, BlockScoutConfigurationProblem
from .mocks import sourcify_safe_metadata


class TestBlockscoutClient(TestCase):
    def test_blockscout_client(self):
        with self.assertRaises(BlockScoutConfigurationProblem):
            BlockscoutClient(EthereumNetwork.MAINNET)

        blockscout_client = BlockscoutClient(EthereumNetwork.XDAI)
        safe_master_copy_abi = sourcify_safe_metadata['output']['abi']
        safe_master_copy_address = '0x6851D6fDFAfD08c0295C392436245E5bc78B0185'
        contract_metadata = blockscout_client.get_contract_metadata(safe_master_copy_address)
        self.assertEqual(contract_metadata.name, 'GnosisSafe')
        self.assertEqual(contract_metadata.abi, safe_master_copy_abi)
        random_address = '0xaE32496491b53841efb51829d6f886387708F99a'
        self.assertIsNone(blockscout_client.get_contract_metadata(random_address))



# flake8: noqa F401
from .blockscout_client import (BlockscoutClient, BlockscoutClientException,
                                BlockScoutConfigurationProblem)
from .contract_metadata import ContractMetadata
from .etherscan_client import (EtherscanClient,
                               EtherscanClientConfigurationProblem,
                               EtherscanClientException,
                               EtherscanRateLimitError)
from .sourcify import Sourcify



AAVE_ATOKEN_ABI = [
    {'inputs': [{'internalType': 'contract ILendingPool',
                 'name': 'pool',
                 'type': 'address'},
                {'internalType': 'address',
                 'name': 'underlyingAssetAddress',
                 'type': 'address'},
                {'internalType': 'address',
                 'name': 'reserveTreasuryAddress',
                 'type': 'address'},
                {'internalType': 'string', 'name': 'tokenName', 'type': 'string'},
                {'internalType': 'string', 'name': 'tokenSymbol', 'type': 'string'},
                {'internalType': 'address',
                 'name': 'incentivesController',
                 'type': 'address'}],
     'stateMutability': 'nonpayable',
     'type': 'constructor'},
    {'anonymous': False,
     'inputs': [{'indexed': True,
                 'internalType': 'address',
                 'name': 'owner',
                 'type': 'address'},
                {'indexed': True,
                 'internalType': 'address',
                 'name': 'spender',
                 'type': 'address'},
                {'indexed': False,
                 'internalType': 'uint256',
                 'name': 'value',
                 'type': 'uint256'}],
     'name': 'Approval',
     'type': 'event'},
    {'anonymous': False,
     'inputs': [{'indexed': True,
                 'internalType': 'address',
                 'name': 'from',
                 'type': 'address'},
                {'indexed': True,
                 'internalType': 'address',
                 'name': 'to',
                 'type': 'address'},
                {'indexed': False,
                 'internalType': 'uint256',
                 'name': 'value',
                 'type': 'uint256'},
                {'indexed': False,
                 'internalType': 'uint256',
                 'name': 'index',
                 'type': 'uint256'}],
     'name': 'BalanceTransfer',
     'type': 'event'},
    {'anonymous': False,
     'inputs': [{'indexed': True,
                 'internalType': 'address',
                 'name': 'from',
                 'type': 'address'},
                {'indexed': True,
                 'internalType': 'address',
                 'name': 'target',
                 'type': 'address'},
                {'indexed': False,
                 'internalType': 'uint256',
                 'name': 'value',
                 'type': 'uint256'},
                {'indexed': False,
                 'internalType': 'uint256',
                 'name': 'index',
                 'type': 'uint256'}],
     'name': 'Burn',
     'type': 'event'},
    {'anonymous': False,
     'inputs': [{'indexed': True,
                 'internalType': 'address',
                 'name': 'underlyingAsset',
                 'type': 'address'},
                {'indexed': True,
                 'internalType': 'address',
                 'name': 'pool',
                 'type': 'address'},
                {'indexed': False,
                 'internalType': 'address',
                 'name': 'treasury',
                 'type': 'address'},
                {'indexed': False,
                 'internalType': 'address',
                 'name': 'incentivesController',
                 'type': 'address'},
                {'indexed': False,
                 'internalType': 'uint8',
                 'name': 'aTokenDecimals',
                 'type': 'uint8'},
                {'indexed': False,
                 'internalType': 'string',
                 'name': 'aTokenName',
                 'type': 'string'},
                {'indexed': False,
                 'internalType': 'string',
                 'name': 'aTokenSymbol',
                 'type': 'string'},
                {'indexed': False,
                 'internalType': 'bytes',
                 'name': 'params',
                 'type': 'bytes'}],
     'name': 'Initialized',
     'type': 'event'},
    {'anonymous': False,
     'inputs': [{'indexed': True,
                 'internalType': 'address',
                 'name': 'from',
                 'type': 'address'},
                {'indexed': False,
                 'internalType': 'uint256',
                 'name': 'value',
                 'type': 'uint256'},
                {'indexed': False,
                 'internalType': 'uint256',
                 'name': 'index',
                 'type': 'uint256'}],
     'name': 'Mint',
     'type': 'event'},
    {'anonymous': False,
     'inputs': [{'indexed': True,
                 'internalType': 'address',
                 'name': 'from',
                 'type': 'address'},
                {'indexed': True,
                 'internalType': 'address',
                 'name': 'to',
                 'type': 'address'},
                {'indexed': False,
                 'internalType': 'uint256',
                 'name': 'value',
                 'type': 'uint256'}],
     'name': 'Transfer',
     'type': 'event'},
    {'inputs': [],
     'name': 'ATOKEN_REVISION',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [],
     'name': 'DOMAIN_SEPARATOR',
     'outputs': [{'internalType': 'bytes32', 'name': '', 'type': 'bytes32'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [],
     'name': 'EIP712_REVISION',
     'outputs': [{'internalType': 'bytes', 'name': '', 'type': 'bytes'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [],
     'name': 'PERMIT_TYPEHASH',
     'outputs': [{'internalType': 'bytes32', 'name': '', 'type': 'bytes32'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [],
     'name': 'POOL',
     'outputs': [{'internalType': 'contract ILendingPool',
                  'name': '',
                  'type': 'address'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [],
     'name': 'RESERVE_TREASURY_ADDRESS',
     'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [],
     'name': 'UINT_MAX_VALUE',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [],
     'name': 'UNDERLYING_ASSET_ADDRESS',
     'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
     'name': '_nonces',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [{'internalType': 'address', 'name': 'owner', 'type': 'address'},
                {'internalType': 'address', 'name': 'spender', 'type': 'address'}],
     'name': 'allowance',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [{'internalType': 'address', 'name': 'spender', 'type': 'address'},
                {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'}],
     'name': 'approve',
     'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'inputs': [{'internalType': 'address', 'name': 'user', 'type': 'address'}],
     'name': 'balanceOf',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [{'internalType': 'address', 'name': 'user', 'type': 'address'},
                {'internalType': 'address',
                 'name': 'receiverOfUnderlying',
                 'type': 'address'},
                {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'},
                {'internalType': 'uint256', 'name': 'index', 'type': 'uint256'}],
     'name': 'burn',
     'outputs': [],
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'inputs': [],
     'name': 'decimals',
     'outputs': [{'internalType': 'uint8', 'name': '', 'type': 'uint8'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [{'internalType': 'address', 'name': 'spender', 'type': 'address'},
                {'internalType': 'uint256', 'name': 'subtractedValue', 'type': 'uint256'}],
     'name': 'decreaseAllowance',
     'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'inputs': [],
     'name': 'getIncentivesController',
     'outputs': [{'internalType': 'contract IAaveIncentivesController',
                  'name': '',
                  'type': 'address'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [{'internalType': 'address', 'name': 'user', 'type': 'address'}],
     'name': 'getScaledUserBalanceAndSupply',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'},
                 {'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [{'internalType': 'address', 'name': 'spender', 'type': 'address'},
                {'internalType': 'uint256', 'name': 'addedValue', 'type': 'uint256'}],
     'name': 'increaseAllowance',
     'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'inputs': [{'internalType': 'uint8',
                 'name': 'underlyingAssetDecimals',
                 'type': 'uint8'},
                {'internalType': 'string', 'name': 'tokenName', 'type': 'string'},
                {'internalType': 'string', 'name': 'tokenSymbol', 'type': 'string'}],
     'name': 'initialize',
     'outputs': [],
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'inputs': [{'internalType': 'address', 'name': 'user', 'type': 'address'},
                {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'},
                {'internalType': 'uint256', 'name': 'index', 'type': 'uint256'}],
     'name': 'mint',
     'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'inputs': [{'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'},
                {'internalType': 'uint256', 'name': 'index', 'type': 'uint256'}],
     'name': 'mintToTreasury',
     'outputs': [],
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'inputs': [],
     'name': 'name',
     'outputs': [{'internalType': 'string', 'name': '', 'type': 'string'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [{'internalType': 'address', 'name': 'owner', 'type': 'address'},
                {'internalType': 'address', 'name': 'spender', 'type': 'address'},
                {'internalType': 'uint256', 'name': 'value', 'type': 'uint256'},
                {'internalType': 'uint256', 'name': 'deadline', 'type': 'uint256'},
                {'internalType': 'uint8', 'name': 'v', 'type': 'uint8'},
                {'internalType': 'bytes32', 'name': 'r', 'type': 'bytes32'},
                {'internalType': 'bytes32', 'name': 's', 'type': 'bytes32'}],
     'name': 'permit',
     'outputs': [],
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'inputs': [{'internalType': 'address', 'name': 'user', 'type': 'address'}],
     'name': 'scaledBalanceOf',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [],
     'name': 'scaledTotalSupply',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [],
     'name': 'symbol',
     'outputs': [{'internalType': 'string', 'name': '', 'type': 'string'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [],
     'name': 'totalSupply',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [{'internalType': 'address',
                 'name': 'recipient',
                 'type': 'address'},
                {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'}],
     'name': 'transfer',
     'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'inputs': [{'internalType': 'address', 'name': 'sender', 'type': 'address'},
                {'internalType': 'address', 'name': 'recipient', 'type': 'address'},
                {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'}],
     'name': 'transferFrom',
     'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'inputs': [{'internalType': 'address', 'name': 'from', 'type': 'address'},
                {'internalType': 'address', 'name': 'to', 'type': 'address'},
                {'internalType': 'uint256', 'name': 'value', 'type': 'uint256'}],
     'name': 'transferOnLiquidation',
     'outputs': [],
     'stateMutability': 'nonpayable',
     'type': 'function'},
    {'inputs': [{'internalType': 'address', 'name': 'target', 'type': 'address'},
                {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'}],
     'name': 'transferUnderlyingTo',
     'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
     'stateMutability': 'nonpayable',
     'type': 'function'}
]



import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

requirements = [
    "cached-property>=1.5; python_version < '3.8'",
    "ethereum>=2.3.2",
    "packaging",
    "py-eth-sig-utils>=0.3.0",
    "typing-extensions==3.10.0.0; python_version < '3.8'",
    "requests>=2",
    "web3>=5.18,<=5.19",
]

extras_require = {
    "django": ["django>=2", "django-filter>=2", "djangorestframework>=2"]
}

setup(
    name='yodaplus-py',
    version='0.9.1',
    packages=find_packages(),
    package_data={'gnosis': ['py.typed']},
    install_requires=requirements,
    include_package_data=True,
    extras_require=extras_require,
    python_requires='>=3.7',
    license='MIT License',
    description='Yodaplus libraries for Python Projects',
    long_description=README,
    url='https://github.com/yodaplus/gnosis-py',
    author='Yodaplus',
    author_email='hello@yodaplus.com',
    keywords=['xdc', 'django', 'rest', 'yplusvault'],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



import os

import pytest
import requests
from eth_account.signers.local import LocalAccount
from web3 import Web3
from web3.contract import Contract

from ..contracts import get_example_erc20_contract


def just_test_if_mainnet_node() -> str:
    mainnet_node_url = os.environ.get('ETHEREUM_MAINNET_NODE')
    if not mainnet_node_url:
        pytest.skip('Mainnet node not defined, cannot test oracles', allow_module_level=True)
    else:
        try:
            if not requests.post(mainnet_node_url, timeout=5,
                                 json={'jsonrpc': '2.0',
                                       'method': 'eth_blockNumber',
                                       'params': [],
                                       'id': 1}).ok:
                pytest.skip('Cannot connect to mainnet node', allow_module_level=True)
        except IOError:
            pytest.skip('Problem connecting to the mainnet node', allow_module_level=True)
    return mainnet_node_url


def send_tx(w3: Web3, tx, account: LocalAccount) -> bytes:
    tx['from'] = account.address
    if 'nonce' not in tx:
        tx['nonce'] = w3.eth.get_transaction_count(account.address, block_identifier='pending')

    if 'gasPrice' not in tx:
        tx['gasPrice'] = w3.eth.gas_price

    if 'gas' not in tx:
        tx['gas'] = w3.eth.estimate_gas(tx)
    else:
        tx['gas'] *= 2

    signed_tx = account.sign_transaction(tx)
    tx_hash = w3.eth.send_raw_transaction(bytes(signed_tx.rawTransaction))
    tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
    assert tx_receipt.status == 1, 'Error with tx %s - %s' % (tx_hash.hex(), tx)
    return tx_hash


def deploy_example_erc20(w3: Web3, amount: int, owner: str, deployer: str = None,
                         account: LocalAccount = None) -> Contract:
    return deploy_erc20(w3, 'Uxio', 'UXI', owner, amount, deployer=deployer, account=account)


def deploy_erc20(w3: Web3, name: str, symbol: str, owner: str, amount: int, decimals: int = 18, deployer: str = None,
                 account: LocalAccount = None) -> Contract:
    if account:
        erc20_contract = get_example_erc20_contract(w3)
        tx = erc20_contract.constructor(name, symbol, decimals, owner, amount).buildTransaction({
            'nonce': w3.eth.get_transaction_count(account.address, block_identifier='pending')
        })
        signed_tx = account.sign_transaction(tx)
        tx_hash = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
    else:
        deployer = deployer or w3.eth.accounts[0]
        erc20_contract = get_example_erc20_contract(w3)
        tx_hash = erc20_contract.constructor(name, symbol, decimals, owner, amount).transact({'from': deployer})

    tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
    erc20_address = tx_receipt.contractAddress
    deployed_erc20 = get_example_erc20_contract(w3, erc20_address)
    assert deployed_erc20.functions.balanceOf(owner).call() == amount
    return deployed_erc20



ZERION_TOKEN_ADAPTER_ABI = [
    {'inputs': [{'internalType': 'address', 'name': 'token', 'type': 'address'}],
     'name': 'getComponents',
     'outputs': [{'components': [{'internalType': 'address',
                                  'name': 'token',
                                  'type': 'address'},
                                 {'internalType': 'string', 'name': 'tokenType', 'type': 'string'},
                                 {'internalType': 'uint256', 'name': 'rate', 'type': 'uint256'}],
                  'internalType': 'struct Component[]',
                  'name': '',
                  'type': 'tuple[]'}],
     'stateMutability': 'view',
     'type': 'function'},
    {'inputs': [{'internalType': 'address', 'name': 'token', 'type': 'address'}],
     'name': 'getMetadata',
     'outputs': [{'components': [{'internalType': 'address',
                                  'name': 'token',
                                  'type': 'address'},
                                 {'internalType': 'string', 'name': 'name', 'type': 'string'},
                                 {'internalType': 'string', 'name': 'symbol', 'type': 'string'},
                                 {'internalType': 'uint8', 'name': 'decimals', 'type': 'uint8'}],
                  'internalType': 'struct TokenMetadata',
                  'name': '',
                  'type': 'tuple'}],
     'stateMutability': 'view',
     'type': 'function'}
]



import logging

from django.test import TestCase

from eth_abi import encode_single
from eth_abi.packed import encode_single_packed
from eth_account import Account
from eth_account.messages import defunct_hash_message
from hexbytes import HexBytes
from web3 import Web3

from ...eth.tests.ethereum_test_case import EthereumTestCaseMixin
from .. import Safe
from ..safe_signature import (SafeSignature, SafeSignatureApprovedHash,
                              SafeSignatureContract, SafeSignatureEOA,
                              SafeSignatureEthSign, SafeSignatureType)
from .safe_test_case import SafeTestCaseMixin

logger = logging.getLogger(__name__)


class TestSafeSignature(EthereumTestCaseMixin, TestCase):
    def test_contract_signature(self):
        owner = '0x05c85Ab5B09Eb8A55020d72daf6091E04e264af9'
        safe_tx_hash = HexBytes('0x4c9577d1b1b8dec52329a983ae26238b65f74b7dd9fb28d74ad9548e92aaf196')
        signature = HexBytes('0x00000000000000000000000005c85ab5b09eb8a55020d72daf6091e04e264af900000000000000000000000'
                             '0000000000000000000000000000000000000000000')
        safe_signature = SafeSignatureContract(signature, safe_tx_hash, b'')
        self.assertEqual(safe_signature.owner, owner)
        self.assertEqual(safe_signature.signature_type, SafeSignatureType.CONTRACT_SIGNATURE)
        self.assertTrue(str(safe_signature))  # Test __str__
        self.assertFalse(safe_signature.is_valid(self.ethereum_client))

    def test_approved_hash_signature(self):
        owner = '0x05c85Ab5B09Eb8A55020d72daf6091E04e264af9'
        safe_tx_hash = HexBytes('0x4c9577d1b1b8dec52329a983ae26238b65f74b7dd9fb28d74ad9548e92aaf196')
        signature = HexBytes('0x00000000000000000000000005c85ab5b09eb8a55020d72daf6091e04e264af900000000000000000000000'
                             '0000000000000000000000000000000000000000001')
        safe_signature = SafeSignatureApprovedHash(signature, safe_tx_hash)
        self.assertEqual(safe_signature.owner, owner)
        self.assertEqual(safe_signature.signature_type, SafeSignatureType.APPROVED_HASH)
        self.assertTrue(str(safe_signature))  # Test __str__

    def test_approved_hash_signature_build(self):
        owner = Account.create().address
        safe_tx_hash = Web3.keccak(text='random')
        safe_signature = SafeSignatureApprovedHash.build_for_owner(owner, safe_tx_hash)
        self.assertEqual(len(safe_signature.signature), 65)
        self.assertEqual(safe_signature.signature_type, SafeSignatureType.APPROVED_HASH)
        self.assertEqual(safe_signature.owner, owner)

        # Check parse signature get the same result
        safe_signature = SafeSignature.parse_signature(safe_signature.signature, safe_tx_hash)[0]
        self.assertEqual(safe_signature.signature_type, SafeSignatureType.APPROVED_HASH)
        self.assertEqual(safe_signature.owner, owner)

    def test_eth_sign_signature(self):
        account = Account.create()
        owner = account.address
        safe_tx_hash = HexBytes('0x123477d1b1b8dec52329a983ae26238b65f74b7dd9fb28d74ad9548e92aaf195')
        message = defunct_hash_message(primitive=safe_tx_hash)
        signature = account.signHash(message)['signature']
        signature = signature[:64] + HexBytes(signature[64] + 4)  # Add 4 to v
        safe_signature = SafeSignatureEthSign(signature, safe_tx_hash)
        self.assertEqual(safe_signature.owner, owner)
        self.assertEqual(safe_signature.signature_type, SafeSignatureType.ETH_SIGN)
        self.assertTrue(safe_signature.is_valid())
        self.assertTrue(str(safe_signature))  # Test __str__

    def test_eoa_signature(self):
        owner = '0xADb7CB706e9A1bd9F96a397da340bF34a9984E1E'
        safe_tx_hash = HexBytes('0x4c9577d1b1b8dec52329a983ae26238b65f74b7dd9fb28d74ad9548e92aaf196')
        signature = HexBytes('0x5dccf9f1375cee56331a67867a0b05a828894c2b76dee84546ec663997d7548257cb2d606087ee7590b966e'
                             '958d97a65528ae941a0f7e5050949f618629509c81b')
        safe_signature = SafeSignatureEOA(signature, safe_tx_hash)
        self.assertEqual(safe_signature.owner, owner)
        self.assertEqual(safe_signature.signature_type, SafeSignatureType.EOA)
        self.assertTrue(safe_signature.is_valid())

        account = Account.create()
        owner = account.address
        safe_tx_hash = HexBytes('0x123477d1b1b8dec52329a983ae26238b65f74b7dd9fb28d74ad9548e92aaf195')
        signature = account.signHash(safe_tx_hash)['signature']
        safe_signature = SafeSignatureEOA(signature, safe_tx_hash)
        self.assertEqual(safe_signature.owner, owner)
        self.assertEqual(safe_signature.signature_type, SafeSignatureType.EOA)
        self.assertTrue(safe_signature.is_valid())
        self.assertTrue(str(safe_signature))  # Test __str__

    def test_defunct_hash_message(self):
        safe_tx_hash = HexBytes('0x4c9577d1b1b8dec52329a983ae26238b65f74b7dd9fb28d74ad9548e92aaf196')
        ethereum_signed_message = '\x19Ethereum Signed Message:\n32'
        encoded_message = encode_single_packed('(string,bytes32)',
                                               (ethereum_signed_message, HexBytes(safe_tx_hash)))
        encoded_hash = Web3.keccak(encoded_message)
        self.assertEqual(encoded_hash, defunct_hash_message(primitive=safe_tx_hash))

    def test_parse_signature(self):
        owner_1 = '0x05c85Ab5B09Eb8A55020d72daf6091E04e264af9'
        owner_2 = '0xADb7CB706e9A1bd9F96a397da340bF34a9984E1E'
        safe_tx_hash = HexBytes('0x4c9577d1b1b8dec52329a983ae26238b65f74b7dd9fb28d74ad9548e92aaf196')
        signatures = HexBytes('0x00000000000000000000000005c85ab5b09eb8a55020d72daf6091e04e264af90000000000000000000000'
                              '000000000000000000000000000000000000000000015dccf9f1375cee56331a67867a0b05a828894c2b76de'
                              'e84546ec663997d7548257cb2d606087ee7590b966e958d97a65528ae941a0f7e5050949f618629509c81b')

        s1, s2 = SafeSignature.parse_signature(signatures, safe_tx_hash)
        self.assertEqual(s1.owner, owner_1)
        self.assertEqual(s1.signature_type, SafeSignatureType.APPROVED_HASH)
        self.assertIsInstance(s1, SafeSignatureApprovedHash)
        self.assertEqual(s2.signature_type, SafeSignatureType.EOA)
        self.assertEqual(s2.owner, owner_2)
        self.assertTrue(s2.is_valid())
        self.assertIsInstance(s2, SafeSignatureEOA)

    def test_parse_signature_with_trailing_zeroes(self):
        signatures = HexBytes('0x9173a0b0895c28cdc82729f0a66e3645820a2e395cadad588463ab483e42b3df6e8b05d258e58c5255136'
                              '0e032733b507f731001cce6b5391b7750439c8cffc020000000000000000000000000dc5299b629ef24fdec'
                              'fbb240c00fc79fabb9cf9700000000000000000000000000000000000000000000000000000000000000000'
                              '1000000000000000000000000000000000000000000000000000000000000')
        parsed_signatures = SafeSignature.parse_signature(signatures, '')
        self.assertEqual(len(parsed_signatures), 2)
        self.assertEqual(parsed_signatures[0].signature_type, SafeSignatureType.ETH_SIGN)
        self.assertEqual(parsed_signatures[1].signature_type, SafeSignatureType.APPROVED_HASH)

    def test_parse_signature_empty(self):
        safe_tx_hash = Web3.keccak(text='Legoshi')
        for value in (b'', '', None):
            self.assertEqual(SafeSignature.parse_signature(value, safe_tx_hash), [])


class TestSafeContractSignature(SafeTestCaseMixin, TestCase):
    def test_contract_signature(self):
        owner_1 = self.ethereum_test_account
        deployed_safe = self.deploy_test_safe(owners=[owner_1.address],
                                              initial_funding_wei=Web3.toWei(0.01, 'ether'))
        safe = Safe(deployed_safe.safe_address, self.ethereum_client)
        safe_contract = safe.get_contract()
        safe_tx_hash = Web3.keccak(text='test')
        signature_r = HexBytes(safe.address.replace('0x', '').rjust(64, '0'))
        signature_s = HexBytes('0' * 62 + '41')  # Position of end of signature `0x41 == 65`
        signature_v = HexBytes('00')
        contract_signature = encode_single('bytes', b'')
        signature = signature_r + signature_s + signature_v + contract_signature

        safe_signature = SafeSignature.parse_signature(signature, safe_tx_hash)[0]
        self.assertFalse(safe_signature.is_valid(self.ethereum_client, None))

        # Check with previously signedMessage
        tx = safe_contract.functions.signMessage(
            safe_tx_hash
        ).buildTransaction({'from': safe.address})
        safe_tx = safe.build_multisig_tx(safe.address, 0, tx['data'])
        safe_tx.sign(owner_1.key)
        safe_tx.execute(owner_1.key)

        safe_signature = SafeSignature.parse_signature(signature, safe_tx_hash)[0]
        self.assertTrue(safe_signature.is_valid(self.ethereum_client, None))
        self.assertIsInstance(safe_signature, SafeSignatureContract)

        # Check with crafted signature
        safe_tx_hash_2 = Web3.keccak(text='test2')
        safe_signature = SafeSignature.parse_signature(signature, safe_tx_hash_2)[0]
        self.assertFalse(safe_signature.is_valid(self.ethereum_client, None))

        safe_tx_hash_2_message_hash = safe_contract.functions.getMessageHash(safe_tx_hash_2).call()
        contract_signature = owner_1.signHash(safe_tx_hash_2_message_hash)['signature']
        encoded_contract_signature = encode_single('bytes', contract_signature)  # It will add size of bytes
        # `32` bytes with the abi encoded size of array. 65 bytes will be padded to next multiple of 32 -> 96
        # 96 - 65 = `31`
        self.assertEqual(len(encoded_contract_signature), len(contract_signature) + 32 + 31)
        crafted_signature = signature_r + signature_s + signature_v + encoded_contract_signature
        safe_signature = SafeSignature.parse_signature(crafted_signature, safe_tx_hash_2)[0]
        self.assertEqual(contract_signature, safe_signature.contract_signature)
        self.assertTrue(safe_signature.is_valid(self.ethereum_client, None))

    def test_contract_multiple_signatures(self):
        """
        Test decode of multiple `CONTRACT_SIGNATURE` together
        """
        owner_1 = self.ethereum_test_account
        deployed_safe = self.deploy_test_safe(owners=[owner_1.address],
                                              initial_funding_wei=Web3.toWei(0.01, 'ether'))
        safe = Safe(deployed_safe.safe_address, self.ethereum_client)
        safe_contract = safe.get_contract()
        safe_tx_hash = Web3.keccak(text='test')

        tx = safe_contract.functions.signMessage(
            safe_tx_hash
        ).buildTransaction({'from': safe.address})
        safe_tx = safe.build_multisig_tx(safe.address, 0, tx['data'])
        safe_tx.sign(owner_1.key)
        safe_tx.execute(owner_1.key)

        # Check multiple signatures. In this case we reuse signatures for the same owner, it won't make sense
        # in real life
        signature_r_1 = HexBytes(safe.address.replace('0x', '').rjust(64, '0'))
        signature_s_1 = HexBytes('0' * 62 + '82')  # Position of end of signature `0x82 == (65 * 2)`
        signature_v_1 = HexBytes('00')
        contract_signature_1 = b''
        encoded_contract_signature_1 = encode_single('bytes', contract_signature_1)

        signature_r_2 = HexBytes(safe.address.replace('0x', '').rjust(64, '0'))
        signature_s_2 = HexBytes('0' * 62 + 'c2')  # Position of end of signature `0xc2 == (65 * 2) + 64`
        signature_v_2 = HexBytes('00')
        safe_tx_hash_message_hash = safe_contract.functions.getMessageHash(safe_tx_hash).call()
        contract_signature_2 = owner_1.signHash(safe_tx_hash_message_hash)['signature']
        encoded_contract_signature_2 = encode_single('bytes', contract_signature_2)  # It will add size of bytes

        signature = (signature_r_1 + signature_s_1 + signature_v_1
                     + signature_r_2 + signature_s_2 + signature_v_2
                     + encoded_contract_signature_1 + encoded_contract_signature_2)

        count = 0
        for safe_signature, contract_signature in zip(SafeSignature.parse_signature(signature, safe_tx_hash),
                                                      [contract_signature_1, contract_signature_2]):
            self.assertEqual(safe_signature.contract_signature, contract_signature)
            self.assertTrue(safe_signature.is_valid(self.ethereum_client, None))
            self.assertEqual(safe_signature.signature_type, SafeSignatureType.CONTRACT_SIGNATURE)
            # Test exported signature
            exported_signature = SafeSignature.parse_signature(safe_signature.export_signature(), safe_tx_hash)[0]
            self.assertEqual(exported_signature.contract_signature, safe_signature.contract_signature)
            self.assertTrue(exported_signature.is_valid(self.ethereum_client, None))
            count += 1
        self.assertEqual(count, 2)



from enum import Enum
from functools import wraps
from logging import getLogger
from typing import (Any, Dict, Iterable, List, NamedTuple, Optional, Sequence,
                    Tuple, Union, cast)

import eth_abi
import requests
from eth_abi.exceptions import DecodingError
from eth_account import Account
from eth_account.signers.local import LocalAccount
from eth_typing import URI, BlockNumber, ChecksumAddress, Hash32, HexStr
from ethereum.utils import (check_checksum, checksum_encode,
                            mk_contract_address, privtoaddr)
from hexbytes import HexBytes
from web3 import HTTPProvider, Web3
from web3._utils.abi import map_abi_data
from web3._utils.method_formatters import (block_formatter, receipt_formatter,
                                           transaction_formatter)
from web3._utils.normalizers import BASE_RETURN_NORMALIZERS
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.exceptions import (BadFunctionCallOutput, BlockNotFound,
                             TimeExhausted, TransactionNotFound)
from web3.middleware import geth_poa_middleware
from web3.types import (BlockData, BlockIdentifier, FilterParams, LogReceipt,
                        Nonce, ParityBlockTrace, ParityFilterParams,
                        ParityFilterTrace, TxData, TxParams, TxReceipt, Wei)

from .constants import (ERC20_721_TRANSFER_TOPIC, GAS_CALL_DATA_BYTE,
                        GAS_CALL_DATA_ZERO_BYTE, NULL_ADDRESS)
from .contracts import get_erc20_contract, get_erc721_contract
from .typing import BalanceDict, EthereumData, EthereumHash
from .utils import decode_string_or_bytes32

logger = getLogger(__name__)


class EthereumNetwork(Enum):
    UNKNOWN = -1
    OLYMPIC = 0
    MAINNET = 1
    ROPSTEN = 3
    OPTIMISTIC = 10
    RINKEBY = 4
    GOERLI = 5
    KOVAN = 42
    BINANCE = 56
    XDAI = 100
    MATIC = 137
    ENERGY_WEB_CHAIN = 246
    FANTOM = 250
    GANACHE = 1337
    ARBITRUM = 42161
    AVALANCHE = 43114
    VOLTA = 73799
    MUMBAI = 80001
    XINFIN = 50
    APOTHEM = 51
    ARBITRUM_TESTNET = 421611
    default = UNKNOWN

    @classmethod
    def _missing_(cls, value):
        return cls.UNKNOWN


class EthereumClientException(ValueError):
    pass


class TransactionAlreadyImported(EthereumClientException):
    pass


class ReplacementTransactionUnderpriced(EthereumClientException):
    pass


class TransactionQueueLimitReached(EthereumClientException):
    pass


class FromAddressNotFound(EthereumClientException):
    pass


class InvalidNonce(EthereumClientException):
    pass


class NonceTooLow(InvalidNonce):
    pass


class NonceTooHigh(InvalidNonce):
    pass


class InsufficientFunds(EthereumClientException):
    pass


class SenderAccountNotFoundInNode(EthereumClientException):
    pass


class UnknownAccount(EthereumClientException):
    pass


class GasLimitExceeded(EthereumClientException):
    pass


class TransactionGasPriceTooLow(EthereumClientException):
    pass


class ParityTraceDecodeException(EthereumClientException):
    pass


class InvalidERC20Info(EthereumClientException):
    pass


class InvalidERC721Info(EthereumClientException):
    pass


def tx_with_exception_handling(func):
    """
    Parity
        - https://github.com/openethereum/openethereum/blob/main/rpc/src/v1/helpers/errors.rs
    Geth
        - https://github.com/ethereum/go-ethereum/blob/master/core/error.go
        - https://github.com/ethereum/go-ethereum/blob/master/core/tx_pool.go
    Comparison
        - https://gist.github.com/kunal365roy/3c37ac9d1c3aaf31140f7c5faa083932

    :param func:
    :return:
    """
    error_with_exception: Dict[str, Exception] = {
        'Transaction with the same hash was already imported': TransactionAlreadyImported,
        'replacement transaction underpriced': ReplacementTransactionUnderpriced,  # https://github.com/ethereum/go-ethereum/blob/eaccdba4ab310e3fb98edbc4b340b5e7c4d767fd/core/tx_pool.go#L72
        'There is another transaction with same nonce in the queue': ReplacementTransactionUnderpriced,  # https://github.com/openethereum/openethereum/blob/f1dc6821689c7f47d8fd07dfc0a2c5ad557b98ec/crates/rpc/src/v1/helpers/errors.rs#L374
        'There are too many transactions in the queue. Your transaction was dropped due to limit. Try increasing '
        'the fee': TransactionQueueLimitReached,  # https://github.com/openethereum/openethereum/blob/f1dc6821689c7f47d8fd07dfc0a2c5ad557b98ec/crates/rpc/src/v1/helpers/errors.rs#L380
        'txpool is full': TransactionQueueLimitReached,  # https://github.com/ethereum/go-ethereum/blob/eaccdba4ab310e3fb98edbc4b340b5e7c4d767fd/core/tx_pool.go#L68
        'transaction underpriced': TransactionGasPriceTooLow,  # https://github.com/ethereum/go-ethereum/blob/eaccdba4ab310e3fb98edbc4b340b5e7c4d767fd/core/tx_pool.go#L64
        'Transaction gas price is too low': TransactionGasPriceTooLow,  # https://github.com/openethereum/openethereum/blob/f1dc6821689c7f47d8fd07dfc0a2c5ad557b98ec/crates/rpc/src/v1/helpers/errors.rs#L386
        'from not found': FromAddressNotFound,
        'correct nonce': InvalidNonce,
        'nonce too low': NonceTooLow,  # https://github.com/ethereum/go-ethereum/blob/bbfb1e4008a359a8b57ec654330c0e674623e52f/core/error.go#L46
        'nonce too high': NonceTooHigh,  # https://github.com/ethereum/go-ethereum/blob/bbfb1e4008a359a8b57ec654330c0e674623e52f/core/error.go#L46
        'insufficient funds': InsufficientFunds,  # https://github.com/openethereum/openethereum/blob/f1dc6821689c7f47d8fd07dfc0a2c5ad557b98ec/crates/rpc/src/v1/helpers/errors.rs#L389
        "doesn't have enough funds": InsufficientFunds,
        'sender account not recognized': SenderAccountNotFoundInNode,
        'unknown account': UnknownAccount,
        'exceeds block gas limit': GasLimitExceeded,  # Geth
        'exceeds current gas limit': GasLimitExceeded,  # https://github.com/openethereum/openethereum/blob/f1dc6821689c7f47d8fd07dfc0a2c5ad557b98ec/crates/rpc/src/v1/helpers/errors.rs#L392
    }

    @wraps(func)
    def with_exception_handling(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ValueError as exc:
            str_exc = str(exc).lower()
            for reason, custom_exception in error_with_exception.items():
                if reason.lower() in str_exc:
                    raise custom_exception(str(exc)) from exc
            raise exc
    return with_exception_handling


class EthereumTxSent(NamedTuple):
    tx_hash: bytes
    tx: TxParams
    contract_address: Optional[str]


class Erc20Info(NamedTuple):
    name: str
    symbol: str
    decimals: int


class Erc721Info(NamedTuple):
    name: str
    symbol: str


class TokenBalance(NamedTuple):
    token_address: str
    balance: int


class EthereumClientProvider:
    def __new__(cls):
        if not hasattr(cls, 'instance'):
            from django.conf import settings
            cls.instance = EthereumClient(settings.ETHEREUM_NODE_URL)
        return cls.instance


class Erc20Manager:
    """
    Manager for ERC20 operations
    """
    # keccak('Transfer(address,address,uint256)')
    # ddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef
    TRANSFER_TOPIC = HexBytes(ERC20_721_TRANSFER_TOPIC)

    def __init__(self, ethereum_client: 'EthereumClient'):
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self.slow_w3 = ethereum_client.slow_w3
        self.http_session = ethereum_client.http_session

    def decode_logs(self, logs: List[Dict[str, Any]]):
        decoded_logs = []
        for log in logs:
            decoded = self._decode_transfer_log(log['data'], log['topics'])
            if decoded:
                log_copy = dict(log)
                log_copy['args'] = decoded
                decoded_logs.append(log_copy)
        return decoded_logs

    def _decode_transfer_log(self, data: EthereumData, topics: Sequence[bytes]) -> Optional[Dict[str, Any]]:
        topics_len = len(topics)
        if topics_len and topics[0] == self.TRANSFER_TOPIC:
            if topics_len == 1:
                # Not standard Transfer(address from, address to, uint256 unknown)
                # 1 topic (transfer topic)
                _from, to, unknown = eth_abi.decode_abi(['address', 'address', 'uint256'], HexBytes(data))
                return {'from': _from, 'to': to, 'unknown': unknown}
            elif topics_len == 3:
                # ERC20 Transfer(address indexed from, address indexed to, uint256 value)
                # 3 topics (transfer topic + from + to)
                value = eth_abi.decode_single('uint256', HexBytes(data))
                _from, to = [Web3.toChecksumAddress(address) for address
                             in eth_abi.decode_abi(['address', 'address'], b''.join(topics[1:]))]
                return {'from': _from, 'to': to, 'value': value}
            elif topics_len == 4:
                # ERC712 Transfer(address indexed from, address indexed to, uint256 indexed tokenId)
                # 4 topics (transfer topic + from + to + tokenId)
                _from, to, token_id = eth_abi.decode_abi(['address', 'address', 'uint256'], b''.join(topics[1:]))
                _from, to = [Web3.toChecksumAddress(address) for address in (_from, to)]
                return {'from': _from, 'to': to, 'tokenId': token_id}
        return None

    def get_balance(self, address: str, token_address: str) -> int:
        """
        Get balance of address for `erc20_address`

        :param address: owner address
        :param token_address: erc20 token address
        :return: balance
        """
        return get_erc20_contract(self.w3, token_address).functions.balanceOf(address).call()

    def get_balances(self, address: str, token_addresses: List[str]) -> List[BalanceDict]:
        """
        Get balances for Ether and tokens for an `address`

        :param address: Owner address checksummed
        :param token_addresses: token addresses to check
        :return: ``List[BalanceDict]``
        """
        # Build ether `eth_getBalance` query
        balance_query = {'jsonrpc': '2.0',
                         'method': 'eth_getBalance',
                         'params': [address, 'latest'],
                         'id': 0}
        queries = [balance_query]

        # Build tokens `balanceOf` query
        for i, erc20_address in enumerate(token_addresses):
            queries.append({'jsonrpc': '2.0',
                            'method': 'eth_call',
                            'params': [{'to': erc20_address,  # Balance of
                                        'data': '0x70a08231' + '{:0>64}'.format(address.replace('0x', '').lower())
                                        }, 'latest'],
                            'id': i + 1})
        response = self.http_session.post(self.ethereum_client.ethereum_node_url, json=queries)
        token_addresses_casted = cast(List[Union[Optional[str]]], [None]) + cast(List[Union[Optional[str]]],
                                                                                 token_addresses)
        balances: List[BalanceDict] = []
        for token_address, data in zip(token_addresses_casted, sorted(response.json(), key=lambda x: x['id'])):
            if 'result' not in data:
                balance = 0
            else:
                try:
                    if token_address:
                        balance = int(self.w3.codec.decode_single('uint256',
                                                                  HexBytes(data['result'])))  # Token
                    else:
                        balance = int(data['result'], 16)  # Ether
                except (ValueError, BadFunctionCallOutput, DecodingError):
                    balance = 0
            balances.append({
                'token_address': token_address,
                'balance': balance,
            })
        return balances

    def get_name(self, erc20_address: str) -> str:
        erc20 = get_erc20_contract(self.w3, erc20_address)
        data = erc20.functions.name().buildTransaction({'gas': Wei(0), 'gasPrice': Wei(0)})['data']
        result = self.w3.eth.call({'to': erc20_address, 'data': data})
        return decode_string_or_bytes32(result)

    def get_symbol(self, erc20_address: str) -> str:
        erc20 = get_erc20_contract(self.w3, erc20_address)
        data = erc20.functions.symbol().buildTransaction({'gas': Wei(0), 'gasPrice': Wei(0)})['data']
        result = self.w3.eth.call({'to': erc20_address, 'data': data})
        return decode_string_or_bytes32(result)

    def get_decimals(self, erc20_address: str) -> int:
        erc20 = get_erc20_contract(self.w3, erc20_address)
        return erc20.functions.decimals().call()

    def get_info(self, erc20_address: str) -> Erc20Info:
        """
        Get erc20 information (`name`, `symbol` and `decimals`). Use batching to get
        all info in the same request.

        :param erc20_address:
        :return: Erc20Info
        :raises: InvalidERC20Info
        """
        erc20 = get_erc20_contract(self.w3, erc20_address)
        params: TxParams = {'gas': Wei(0), 'gasPrice': Wei(0)}  # Prevent executing tx, we are just interested on `data`
        datas = [erc20.functions.name().buildTransaction(params)['data'],
                 erc20.functions.symbol().buildTransaction(params)['data'],
                 erc20.functions.decimals().buildTransaction(params)['data']]
        payload = [{'id': i, 'jsonrpc': '2.0', 'method': 'eth_call',
                    'params': [{'to': erc20_address, 'data': data}, 'latest']}
                   for i, data in enumerate(datas)]
        response = self.http_session.post(self.ethereum_client.ethereum_node_url, json=payload)
        if not response.ok:
            raise InvalidERC20Info(response.content)
        try:
            response_json = sorted(response.json(), key=lambda x: x['id'])
            errors = [r['error'] for r in response_json if 'error' in r]
            if errors:
                raise InvalidERC20Info(f'{erc20_address} - {errors}')
            results = [HexBytes(r['result']) for r in response_json]
            name = decode_string_or_bytes32(results[0])
            symbol = decode_string_or_bytes32(results[1])
            decimals = self.ethereum_client.w3.codec.decode_single('uint8', results[2])
            return Erc20Info(name, symbol, decimals)
        except (ValueError, BadFunctionCallOutput, DecodingError) as e:
            raise InvalidERC20Info from e

    def get_total_transfer_history(self, addresses: Optional[Sequence[ChecksumAddress]] = None,
                                   from_block: BlockIdentifier = BlockNumber(0),
                                   to_block: Optional[BlockIdentifier] = None,
                                   token_address: Optional[ChecksumAddress] = None) -> List[Dict[str, Any]]:
        """
        Get events for erc20 and erc721 transfers from and to an `address`. We decode it manually.
        Example of an erc20 event:

        .. code-block:: python

            {'logIndex': 0,
             'transactionIndex': 0,
             'transactionHash': HexBytes('0x4d0f25313603e554e3b040667f7f391982babbd195c7ae57a8c84048189f7794'),
             'blockHash': HexBytes('0x90fa67d848a0eaf3be625235dae28815389f5292d4465c48d1139f0c207f8d42'),
             'blockNumber': 791,
             'address': '0xf7d0Bd47BF3214494E7F5B40E392A25cb4788620',
             'data': '0x000000000000000000000000000000000000000000000000002001f716742000',
             'topics': [HexBytes('0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'),
              HexBytes('0x000000000000000000000000f5984365fca2e3bc7d2e020abb2c701df9070eb7'),
              HexBytes('0x0000000000000000000000001df62f291b2e969fb0849d99d9ce41e2f137006e')],
             'type': 'mined'
             'args': {'from': '0xf5984365FcA2e3bc7D2E020AbB2c701DF9070eB7',
                      'to': '0x1dF62f291b2E969fB0849d99D9Ce41e2F137006e',
                      'value': 9009360000000000
                     }
            }
            An example of an erc721 event
            {'address': '0x6631FcbB50677DfC6c02CCDcc03a8f68Db427a64',
             'blockHash': HexBytes('0x95c71c6c9373e9a8ca2c767dda1cd5083eb6addcce36fc216c9e1f458d6970f9'),
             'blockNumber': 5341681,
             'data': '0x',
             'logIndex': 0,
             'removed': False,
             'topics': [HexBytes('0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'),
              HexBytes('0x0000000000000000000000000000000000000000000000000000000000000000'),
              HexBytes('0x000000000000000000000000b5239c032ab9fb5abfc3903e770a4b6a9095542c'),
              HexBytes('0x0000000000000000000000000000000000000000000000000000000000000063')],
             'transactionHash': HexBytes('0xce8c8af0503e6f8a421345c10cdf92834c95186916a3f5b1437d2bba63d2db9e'),
             'transactionIndex': 0,
             'transactionLogIndex': '0x0',
             'type': 'mined',
             'args': {'from': '0x0000000000000000000000000000000000000000',
                      'to': '0xb5239C032AB9fB5aBFc3903e770A4B6a9095542C',
                      'tokenId': 99
                     }
             }
            An example of unknown transfer event (no indexed parts), could be a ERC20 or ERC721 transfer:
            {'address': '0x6631FcbB50677DfC6c02CCDcc03a8f68Db427a64',
             'blockHash': HexBytes('0x95c71c6c9373e9a8ca2c767dda1cd5083eb6addcce36fc216c9e1f458d6970f9'),
             'blockNumber': 5341681,
             'data': '0x',
             'logIndex': 0,
             'removed': False,
             'topics': [HexBytes('0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'),
              HexBytes('0x0000000000000000000000000000000000000000000000000000000000000000'),
              HexBytes('0x000000000000000000000000b5239c032ab9fb5abfc3903e770a4b6a9095542c'),
              HexBytes('0x0000000000000000000000000000000000000000000000000000000000000063')],
             'transactionHash': HexBytes('0xce8c8af0503e6f8a421345c10cdf92834c95186916a3f5b1437d2bba63d2db9e'),
             'transactionIndex': 0,
             'transactionLogIndex': '0x0',
             'type': 'mined',
             'args': {'from': '0x0000000000000000000000000000000000000000',
                      'to': '0xb5239C032AB9fB5aBFc3903e770A4B6a9095542C',
                      'unknown': 99
                     }
             }

        :param addresses: Search events `from` and `to` these `addresses`. If not, every transfer event within the
            range will be retrieved
        :param from_block: Block to start querying from
        :param to_block: Block to stop querying from
        :param token_address: Address of the token
        :return: List of events sorted by blockNumber
        """
        topic_0 = self.TRANSFER_TOPIC.hex()
        if addresses:
            addresses_encoded = [HexBytes(eth_abi.encode_single('address', address)).hex() for address in addresses]
            # Topics for transfer `to` and `from` an address
            all_topics = [
                [topic_0, addresses_encoded],  # Topics from
                [topic_0, None, addresses_encoded],  # Topics to
            ]
        else:
            all_topics = [
                [topic_0]  # All transfer events
            ]
        parameters: FilterParams = {'fromBlock': from_block}
        if to_block:
            parameters['toBlock'] = to_block
        if token_address:
            parameters['address'] = token_address

        all_events: List[LogReceipt] = []
        # Do the request to `eth_getLogs`
        for topics in all_topics:
            parameters['topics'] = topics
            all_events.extend(self.slow_w3.eth.get_logs(parameters))

        # Decode events. Just pick valid ERC20 Transfer events (ERC721 `Transfer` has the same signature)
        erc20_events = []
        for event in all_events:
            e = LogReceipt(event)  # Convert `AttributeDict` to `Dict`
            e['args'] = self._decode_transfer_log(e['data'], e['topics'])
            if e['args']:
                erc20_events.append(e)
        erc20_events.sort(key=lambda x: x['blockNumber'])
        return erc20_events

    def get_transfer_history(self, from_block: int, to_block: Optional[int] = None,
                             from_address: Optional[str] = None, to_address: Optional[str] = None,
                             token_address: Optional[str] = None) -> List[Dict[str, Any]]:
        """
        DON'T USE, it will fail in some cases until they fix https://github.com/ethereum/web3.py/issues/1351
        Get events for erc20/erc721 transfers. At least one of `from_address`, `to_address` or `token_address` must be
        defined. Example of decoded event:

        .. code-block:: python

            {
                "args": {
                    "from": "0x1Ce67Ea59377A163D47DFFc9BaAB99423BE6EcF1",
                    "to": "0xaE9E15896fd32E59C7d89ce7a95a9352D6ebD70E",
                    "value": 15000000000000000
                },
                "event": "Transfer",
                "logIndex": 42,
                "transactionIndex": 60,
                "transactionHash": "0x71d6d83fef3347bad848e83dfa0ab28296e2953de946ee152ea81c6dfb42d2b3",
                "address": "0xfecA834E7da9D437645b474450688DA9327112a5",
                "blockHash": "0x054de9a496fc7d10303068cbc7ee3e25181a3b26640497859a5e49f0342e7db2",
                "blockNumber": 7265022
            }

        :param from_block: Block to start querying from
        :param to_block: Block to stop querying from
        :param from_address: Address sending the erc20 transfer
        :param to_address: Address receiving the erc20 transfer
        :param token_address: Address of the token
        :return: List of events (decoded)
        :throws: ReadTimeout
        """
        assert from_address or to_address or token_address, 'At least one parameter must be provided'

        erc20 = get_erc20_contract(self.slow_w3)

        argument_filters = {}
        if from_address:
            argument_filters['from'] = from_address
        if to_address:
            argument_filters['to'] = to_address

        return erc20.events.Transfer.createFilter(fromBlock=from_block,
                                                  toBlock=to_block,
                                                  address=token_address,
                                                  argument_filters=argument_filters).get_all_entries()

    def send_tokens(self, to: str, amount: int, erc20_address: str, private_key: str,
                    nonce: Optional[int] = None, gas_price: Optional[int] = None, gas: Optional[int] = None) -> bytes:
        """
        Send tokens to address

        :param to:
        :param amount:
        :param erc20_address:
        :param private_key:
        :param nonce:
        :param gas_price:
        :param gas:
        :return: tx_hash
        """
        erc20 = get_erc20_contract(self.w3, erc20_address)
        account = Account.from_key(private_key)
        tx_options: TxParams = {'from': account.address}
        if nonce:
            tx_options['nonce'] = Nonce(nonce)
        if gas_price:
            tx_options['gasPrice'] = Wei(gas_price)
        if gas:
            tx_options['gas'] = Wei(gas)

        tx = erc20.functions.transfer(to, amount).buildTransaction(tx_options)
        return self.ethereum_client.send_unsigned_transaction(tx, private_key=private_key)


class Erc721Manager:
    # keccak('Transfer(address,address,uint256)')
    # ddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef
    TRANSFER_TOPIC = Erc20Manager.TRANSFER_TOPIC

    def __init__(self, ethereum_client: 'EthereumClient'):
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self.slow_w3 = ethereum_client.slow_w3

    def get_balance(self, address: str, token_address: str) -> int:
        """
        Get balance of address for `erc20_address`

        :param address: owner address
        :param token_address: erc721 token address
        :return: balance
        """
        return get_erc721_contract(self.w3, token_address).functions.balanceOf(address).call()

    def get_balances(self, address: str, token_addresses: List[str]) -> List[TokenBalance]:
        """
        Get balances for tokens for an `address`. If there's a problem with a token_address `0` will be
        returned for balance

        :param address: Owner address checksummed
        :param token_addresses: token addresses to check
        :return:
        """
        balances = self.ethereum_client.batch_call(
            [get_erc721_contract(self.ethereum_client.w3, token_address).functions.balanceOf(address)
             for token_address in token_addresses],
            raise_exception=False
        )
        return [TokenBalance(token_address, 0 if balance is None else balance)
                for (token_address, balance) in zip(token_addresses, balances)]

    def get_info(self, token_address: str) -> Erc721Info:
        """
        Get erc721 information (`name`, `symbol`). Use batching to get
        all info in the same request.

        :param token_address:
        :return: Erc721Info
        """
        erc721_contract = get_erc721_contract(self.w3, token_address)
        try:
            name, symbol = cast(List[str], self.ethereum_client.batch_call([
                erc721_contract.functions.name(),
                erc721_contract.functions.symbol(),
            ]))
            return Erc721Info(name, symbol)
        except (DecodingError, ValueError):  # Not all the ERC721 have metadata
            raise InvalidERC721Info

    def get_owners(self, token_addresses_with_token_ids: Sequence[Tuple[str, int]]) -> List[Optional[str]]:
        """
        :param token_addresses_with_token_ids: Tuple(token_address: str, token_id: int)
        :return: List of owner addresses, `None` if not found
        """
        return cast(List[Optional[str]], self.ethereum_client.batch_call(
            [get_erc721_contract(self.ethereum_client.w3, token_address).functions.ownerOf(token_id)
             for token_address, token_id in token_addresses_with_token_ids],
            raise_exception=False
        ))

    def get_token_uris(self, token_addresses_with_token_ids: Sequence[Tuple[str, int]]) -> List[Optional[str]]:
        """
        :param token_addresses_with_token_ids: Tuple(token_address: str, token_id: int)
        :return: List of token_uris, `None` if not found
        """
        return cast(List[Optional[str]], self.ethereum_client.batch_call(
            [get_erc721_contract(self.ethereum_client.w3, token_address).functions.tokenURI(token_id)
             for token_address, token_id in token_addresses_with_token_ids],
            raise_exception=False
        ))


class ParityManager:
    def __init__(self, ethereum_client: 'EthereumClient'):
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self.slow_w3 = ethereum_client.slow_w3
        self.ethereum_node_url = ethereum_client.ethereum_node_url
        self.http_session = ethereum_client.http_session

    # TODO Test with mock
    def _decode_trace_action(self, action: Dict[str, Any]) -> Dict[str, Any]:
        decoded = {
        }

        # CALL, DELEGATECALL, CREATE or CREATE2
        if 'from' in action:
            decoded['from'] = self.w3.toChecksumAddress(action['from'])
        if 'gas' in action:
            decoded['gas'] = int(action['gas'], 16)
        if 'value' in action:
            decoded['value'] = int(action['value'], 16)

        # CALL or DELEGATECALL
        if 'callType' in action:
            decoded['callType'] = action['callType']
        if 'input' in action:
            decoded['input'] = HexBytes(action['input'])
        if 'to' in action:
            decoded['to'] = self.w3.toChecksumAddress(action['to'])

        # CREATE or CREATE2
        if 'init' in action:
            decoded['init'] = HexBytes(action['init'])

        # SELF-DESTRUCT
        if 'address' in action:
            decoded['address'] = self.w3.toChecksumAddress(action['address'])
        if 'balance' in action:
            decoded['balance'] = int(action['balance'], 16)
        if 'refundAddress' in action:
            decoded['refundAddress'] = self.w3.toChecksumAddress(action['refundAddress'])

        return decoded

    def _decode_trace_result(self, result: Dict[str, Any]) -> Dict[str, Any]:
        decoded: Dict[str, Any] = {
            'gasUsed': int(result['gasUsed'], 16),
        }

        # CALL or DELEGATECALL
        if 'output' in result:
            decoded['output'] = HexBytes(result['output'])

        # CREATE or CREATE2
        if 'code' in result:
            decoded['code'] = HexBytes(result['code'])
        if 'address' in result:
            decoded['address'] = self.w3.toChecksumAddress(result['address'])

        return decoded

    def _decode_traces(self, traces: Sequence[Union[ParityBlockTrace, ParityFilterTrace]]) -> List[Dict[str, Any]]:
        new_traces: List[Dict[str, Any]] = []
        for trace in traces:
            if isinstance(trace, dict):
                trace_copy = trace.copy()
            elif isinstance(trace, AttributeDict):
                trace_copy = trace.__dict__.copy()
            else:
                raise ParityTraceDecodeException('Expected dictionary, but found unexpected trace %s' % trace)
            new_traces.append(trace_copy)
            # Txs with `error` field don't have `result` field
            # Txs with `type=suicide` have `result` field but is `None`
            if 'result' in trace and trace['result']:
                trace_copy['result'] = self._decode_trace_result(trace['result'])
            trace_copy['action'] = self._decode_trace_action(trace['action'])
        return new_traces

    def filter_out_errored_traces(self, internal_txs: Sequence[Dict[str, Any]]) -> Sequence[Dict[str, Any]]:
        """
        Filter out errored transactions (traces that are errored or that have an errored parent)

        :param internal_txs: Traces for the SAME ethereum tx, sorted ascending by `trace_address`
            `sorted(t, key = lambda i: i['traceAddress'])`. It's the default output from methods returning `traces` like
            `trace_block` or `trace_transaction`
        :return: List of not errored traces
        """
        new_list = []
        errored_trace_address: Optional[List[int]] = None
        for internal_tx in internal_txs:
            if internal_tx.get('error') is not None:
                errored_trace_address = internal_tx['traceAddress']
            elif (errored_trace_address is not None
                  and internal_tx['traceAddress'][:len(errored_trace_address)] == errored_trace_address):
                continue
            else:
                new_list.append(internal_tx)
        return new_list

    def get_previous_trace(self, tx_hash: EthereumHash, trace_address: Sequence[int],
                           number_traces: int = 1, skip_delegate_calls: bool = False) -> Optional[Dict[str, Any]]:
        """
        :param tx_hash:
        :param trace_address:
        :param number_traces: Number of traces to skip, by default get the immediately previous one
        :param skip_delegate_calls: If True filter out delegate calls
        :return: Parent trace for a trace
        :raises: ``ValueError`` if tracing is not supported
        """
        if len(trace_address) < number_traces:
            return None

        trace_address = trace_address[:-number_traces]
        traces = reversed(self.trace_transaction(tx_hash))
        for trace in traces:
            if trace_address == trace['traceAddress']:
                if skip_delegate_calls and trace['action'].get('callType') == 'delegatecall':
                    trace_address = trace_address[:-1]
                else:
                    return trace

    def get_next_traces(self, tx_hash: EthereumHash, trace_address: Sequence[int],
                        remove_delegate_calls: bool = False, remove_calls: bool = False) -> List[Dict[str, Any]]:
        """
        :param tx_hash:
        :param trace_address:
        :param remove_delegate_calls: If True remove delegate calls from result
        :param remove_calls: If True remove calls from result
        :return: Children for a trace, E.g. if address is [0, 1] and number_traces = 1, it will return [0, 1, x]
        :raises: ``ValueError`` if tracing is not supported
        """
        trace_address_len = len(trace_address)
        traces = []
        for trace in self.trace_transaction(tx_hash):
            if trace_address_len + 1 == len(trace['traceAddress']) and trace_address == trace['traceAddress'][:-1]:
                if remove_delegate_calls and trace['action'].get('callType') == 'delegatecall':
                    pass
                elif remove_calls and trace['action'].get('callType') == 'call':
                    pass
                else:
                    traces.append(trace)
        return traces

    def trace_block(self, block_identifier: BlockIdentifier) -> List[Dict[str, Any]]:
        try:
            return self._decode_traces(self.slow_w3.parity.trace_block(block_identifier))
        except ParityTraceDecodeException as exc:
            logger.warning('Problem decoding trace: %s - Retrying', exc)
            return self._decode_traces(self.slow_w3.parity.trace_block(block_identifier))

    def trace_blocks(self, block_identifiers: List[BlockIdentifier]) -> List[List[Dict[str, Any]]]:
        if not block_identifiers:
            return []
        payload = [{'id': i, 'jsonrpc': '2.0', 'method': 'trace_block',
                    'params': [hex(block_identifier) if isinstance(block_identifier, int) else block_identifier]}
                   for i, block_identifier in enumerate(block_identifiers)]
        response = self.http_session.post(self.ethereum_node_url, json=payload)
        if not response.ok:
            message = f'Problem calling batch `trace_block` on blocks={block_identifiers} ' \
                      f'status_code={response.status_code} result={response.content}'
            logger.error(message)
            raise ValueError(message)
        results = sorted(response.json(), key=lambda x: x['id'])
        traces = []
        for block_identifier, result in zip(block_identifiers, results):
            if 'result' not in result:
                message = f'Problem calling batch `trace_block` on block={block_identifier}, result={result}'
                logger.error(message)
                raise ValueError(message)
            raw_tx = result['result']
            if raw_tx:
                try:
                    decoded_traces = self._decode_traces(raw_tx)
                except ParityTraceDecodeException as exc:
                    logger.warning('Problem decoding trace: %s - Retrying', exc)
                    decoded_traces = self._decode_traces(raw_tx)
                traces.append(decoded_traces)
            else:
                traces.append([])
        return traces

    def trace_transaction(self, tx_hash: EthereumHash) -> List[Dict[str, Any]]:
        """
        :param tx_hash:
        :return: List of internal txs for `tx_hash`
        """
        try:
            return self._decode_traces(self.slow_w3.parity.trace_transaction(tx_hash))
        except ParityTraceDecodeException as exc:
            logger.warning('Problem decoding trace: %s - Retrying', exc)
            return self._decode_traces(self.slow_w3.parity.trace_transaction(tx_hash))

    def trace_transactions(self, tx_hashes: Sequence[EthereumHash]) -> List[List[Dict[str, Any]]]:
        """
        :param tx_hashes:
        :return: For every `tx_hash` a list of internal txs (in the same order as the `tx_hashes` were provided)
        """
        if not tx_hashes:
            return []
        payload = [{'id': i, 'jsonrpc': '2.0', 'method': 'trace_transaction',
                    'params': [HexBytes(tx_hash).hex()]}
                   for i, tx_hash in enumerate(tx_hashes)]
        response = self.http_session.post(self.ethereum_node_url, json=payload)
        if not response.ok:
            message = f'Problem calling batch `trace_transaction` on tx_hashes={tx_hashes} ' \
                      f'status_code={response.status_code} result={response.content}'
            logger.error(message)
            raise ValueError(message)
        results = sorted(response.json(), key=lambda x: x['id'])
        traces = []
        for result in results:
            raw_tx = result['result']
            if raw_tx:
                try:
                    decoded_traces = self._decode_traces(raw_tx)
                except ParityTraceDecodeException as exc:
                    logger.warning('Problem decoding trace: %s - Retrying', exc)
                    decoded_traces = self._decode_traces(raw_tx)
                traces.append(decoded_traces)
            else:
                traces.append([])
        return traces

    def trace_filter(self, from_block: int = 1, to_block: Optional[int] = None,
                     from_address: Optional[Sequence[ChecksumAddress]] = None,
                     to_address: Optional[Sequence[ChecksumAddress]] = None,
                     after: Optional[int] = None, count: Optional[int] = None) -> List[Dict[str, Any]]:
        """
        Get events using ``trace_filter`` method

        :param from_block: Quantity or Tag - (optional) From this block. `0` is not working, it needs to be `>= 1`
        :param to_block: Quantity or Tag - (optional) To this block.
        :param from_address: Array - (optional) Sent from these addresses.
        :param to_address: Address - (optional) Sent to these addresses.
        :param after: Quantity - (optional) The offset trace number
        :param count: Quantity - (optional) Integer number of traces to display in a batch.
        :return:

        .. code-block:: python

            [
                {
                    "action": {
                        "callType": "call",
                        "from": "0x32be343b94f860124dc4fee278fdcbd38c102d88",
                        "gas": "0x4c40d",
                        "input": "0x",
                        "to": "0x8bbb73bcb5d553b5a556358d27625323fd781d37",
                        "value": "0x3f0650ec47fd240000"
                    },
                    "blockHash": "0x86df301bcdd8248d982dbf039f09faf792684e1aeee99d5b58b77d620008b80f",
                    "blockNumber": 3068183,
                    "result": {
                        "gasUsed": "0x0",
                        "output": "0x"
                    },
                    "subtraces": 0,
                    "traceAddress": [],
                    "transactionHash": "0x3321a7708b1083130bd78da0d62ead9f6683033231617c9d268e2c7e3fa6c104",
                    "transactionPosition": 3,
                    "type": "call"
                },
                {
                    "action": {
                        "from": "0x3b169a0fb55ea0b6bafe54c272b1fe4983742bf7",
                        "gas": "0x49b0b",
                        "init": " ",
                        "value": "0x0"
                    },
                    "blockHash": "0x03f9f64dfeb7807b5df608e6957dd4d521fd71685aac5533451d27f0abe03660",
                    "blockNumber": 3793534,
                    "result": {
                        "address": "0x61a7cc907c47c133d5ff5b685407201951fcbd08",
                        "code": "0x60806040526004361061004c576000357c0100000000000000000000000000000000000000000000000000000000900463ffffffff1680634555d5c91461008b5780635c60da1b146100b6575b73ffffffffffffffffffffffffffffffffffffffff600054163660008037600080366000845af43d6000803e6000811415610086573d6000fd5b3d6000f35b34801561009757600080fd5b506100a061010d565b6040518082815260200191505060405180910390f35b3480156100c257600080fd5b506100cb610116565b604051808273ffffffffffffffffffffffffffffffffffffffff1673ffffffffffffffffffffffffffffffffffffffff16815260200191505060405180910390f35b60006002905090565b60008060009054906101000a900473ffffffffffffffffffffffffffffffffffffffff169050905600a165627a7a7230582007fffd557dfc8c4d2fdf56ba6381a6ce5b65b6260e1492d87f26c6d4f1d041080029",
                        "gasUsed": "0x4683f"
                    },
                    "subtraces": 2,
                    "traceAddress": [],
                    "transactionHash": "0x6c7e8f8778d33d81b29c4bd7526ee50a4cea340d69eed6c89ada4e6fab731789",
                    "transactionPosition": 1,
                    "type": "create"
                },
                {
                    'action': {
                        'address': '0x4440adafbc6c4e45c299451c0eedc7c8b98c14ac',
                        'balance': '0x0',
                        'refundAddress': '0x0000000000000000000000000000000000000000'
                    },
                    'blockHash': '0x8512d367492371edf44ebcbbbd935bc434946dddc2b126cb558df5906012186c',
                    'blockNumber': 7829689,
                    'result': None,
                    'subtraces': 0,
                    'traceAddress': [0, 0, 0, 0, 0, 0],
                    'transactionHash': '0x5f7af6aa390f9f8dd79ee692c37cbde76bb7869768b1bac438b6d176c94f637d',
                    'transactionPosition': 35,
                    'type': 'suicide'
                }
            ]

        """
        assert from_address or to_address, 'You must provide at least `from_address` or `to_address`'
        parameters: ParityFilterParams = {}
        if after:
            parameters['after'] = after
        if count:
            parameters['count'] = count
        if from_block:
            parameters['fromBlock'] = HexStr('0x%x' % from_block)
        if to_block:
            parameters['toBlock'] = HexStr('0x%x' % to_block)
        if from_address:
            parameters['fromAddress'] = from_address
        if to_address:
            parameters['toAddress'] = to_address

        try:
            return self._decode_traces(self.slow_w3.parity.trace_filter(parameters))
        except ParityTraceDecodeException as exc:
            logger.warning('Problem decoding trace: %s - Retrying', exc)
            return self._decode_traces(self.slow_w3.parity.trace_filter(parameters))


class EthereumClient:
    """
    Manage ethereum operations. Uses web3 for the most part, but some other stuff is implemented from scratch.
    Note: If you want to use `pending` state with `Parity`, it must be run with `--pruning=archive` or `--force-sealing`
    """
    NULL_ADDRESS = NULL_ADDRESS

    def __init__(self, ethereum_node_url: URI = URI('http://localhost:8545'), slow_provider_timeout: int = 60):
        self.http_session = self._prepare_http_session()
        self.ethereum_node_url: str = ethereum_node_url
        self.w3_provider = HTTPProvider(self.ethereum_node_url, session=self.http_session)
        self.w3_slow_provider = HTTPProvider(self.ethereum_node_url,
                                             request_kwargs={'timeout': slow_provider_timeout},
                                             session=self.http_session)
        self.w3: Web3 = Web3(self.w3_provider)
        self.slow_w3: Web3 = Web3(self.w3_slow_provider)
        self.erc20: Erc20Manager = Erc20Manager(self)
        self.erc721: Erc721Manager = Erc721Manager(self)
        self.parity: ParityManager = ParityManager(self)
        try:
            if int(self.w3.net.version) != 1:
                self.w3.middleware_onion.inject(geth_poa_middleware, layer=0)
            # For tests using dummy connections (like IPC)
        except (ConnectionError, FileNotFoundError):
            self.w3.middleware_onion.inject(geth_poa_middleware, layer=0)

    def __str__(self):
        return f'EthereumClient for url={self.ethereum_node_url}'

    def _prepare_http_session(self):
        """
        Prepare http session with custom pooling. See:
        https://urllib3.readthedocs.io/en/stable/advanced-usage.html
        https://2.python-requests.org/en/latest/api/#requests.adapters.HTTPAdapter
        https://web3py.readthedocs.io/en/stable/providers.html#httpprovider
        """
        session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(pool_connections=1,  # Doing all the connections to the same url
                                                pool_maxsize=100)  # Do many requests to the same host
        session.mount('http://', adapter)
        session.mount('https://', adapter)
        return session

    @property
    def current_block_number(self):
        return self.w3.eth.block_number

    def deploy_and_initialize_contract(self, deployer_account: LocalAccount,
                                       constructor_data: bytes, initializer_data: bytes = b'',
                                       check_receipt: bool = True):
        contract_address: Optional[ChecksumAddress] = None
        for data in (constructor_data, initializer_data):
            # Because initializer_data is not mandatory
            if data:
                tx: TxParams = {'from': deployer_account.address,
                                'data': data,
                                'gasPrice': self.w3.eth.gas_price,
                                'value': Wei(0),
                                'to': contract_address if contract_address else ''}
                tx['gas'] = self.w3.eth.estimate_gas(tx)
                tx_hash = self.send_unsigned_transaction(tx, private_key=deployer_account.key)
                if check_receipt:
                    tx_receipt = self.get_transaction_receipt(Hash32(tx_hash), timeout=60)
                    assert tx_receipt
                    assert tx_receipt['status']

                if not contract_address:
                    contract_address = ChecksumAddress(checksum_encode(mk_contract_address(tx['from'], tx['nonce'])))

        return EthereumTxSent(tx_hash, tx, contract_address)

    def get_network(self) -> EthereumNetwork:
        """
        Get network name based on the network version id

        :return: The EthereumNetwork enum type
        """
        return EthereumNetwork(int(self.w3.net.version))

    def get_nonce_for_account(self, address: ChecksumAddress, block_identifier: Optional[BlockIdentifier] = 'latest'):
        """
        Get nonce for account. `getTransactionCount` is the only method for what `pending` is currently working
        (Geth and Parity)

        :param address:
        :param block_identifier:
        :return:
        """
        return self.w3.eth.get_transaction_count(address, block_identifier=block_identifier)

    def batch_call_custom(self, payloads: Iterable[Dict[str, Any]],
                          raise_exception: bool = True,
                          block_identifier: Optional[BlockIdentifier] = 'latest') -> List[Optional[Any]]:
        """
        Do batch requests of multiple contract calls

        :param payloads: Iterable of Dictionaries with at least {'data': '<hex-string>',
            'output_type': <solidity-output-type>, 'to': '<checksummed-address>'}. `from` can also be provided and if
            `fn_name` is provided it will be used for debugging purposes
        :param raise_exception: If False, exception will not be raised if there's any problem and instead `None` will
            be returned as the value
        :param block_identifier: `latest` by default
        :return: List with the ABI decoded return values
        :raises: ValueError if raise_exception=True
        """
        if not payloads:
            return []

        queries = []
        for i, payload in enumerate(payloads):
            assert 'data' in payload, '`data` not present'
            assert 'to' in payload, '`to` not present'
            assert 'output_type' in payload, '`output-type` not present'

            query_params = {'to': payload['to'],  # Balance of
                            'data': payload['data']
                            }
            if 'from' in payload:
                query_params['from'] = payload['from']

            queries.append({'jsonrpc': '2.0',
                            'method': 'eth_call',
                            'params': [query_params,
                                       hex(block_identifier) if isinstance(block_identifier,
                                                                           int) else block_identifier],
                            'id': i})

        response = self.http_session.post(self.ethereum_node_url, json=queries)
        if not response.ok:
            raise ConnectionError(f'Error connecting to {self.ethereum_node_url}: {response.text}')

        return_values: List[Optional[Any]] = []
        errors = []
        for payload, result in zip(payloads, sorted(response.json(), key=lambda x: x['id'])):
            if 'error' in result:
                fn_name = payload.get('fn_name', HexBytes(payload['data']).hex())
                errors.append(f'`{fn_name}`: {result["error"]}')
                return_values.append(None)
            else:
                output_type = payload['output_type']
                try:
                    decoded_values = self.w3.codec.decode_abi(output_type, HexBytes(result['result']))
                    normalized_data = map_abi_data(BASE_RETURN_NORMALIZERS, output_type, decoded_values)
                    if len(normalized_data) == 1:
                        return_values.append(normalized_data[0])
                    else:
                        return_values.append(normalized_data)
                except (DecodingError, OverflowError):
                    fn_name = payload.get('fn_name', HexBytes(payload['data']).hex())
                    errors.append(f'`{fn_name}`: DecodingError, cannot decode')
                    return_values.append(None)

        if errors and raise_exception:
            raise ValueError(f'Errors returned {errors}')
        else:
            return return_values

    def batch_call(self, contract_functions: Iterable[ContractFunction], from_address: Optional[str] = None,
                   raise_exception: bool = True,
                   block_identifier: Optional[BlockIdentifier] = 'latest') -> List[Optional[Any]]:
        """
        Do batch requests of multiple contract calls

        :param contract_functions: Iterable of contract functions using web3.py contracts. For instance, a valid
            argument would be [erc20_contract.functions.balanceOf(address), erc20_contract.functions.decimals()]
        :param from_address: Use this address as `from` in every call if provided
        :param block_identifier: `latest` by default
        :param raise_exception: If False, exception will not be raised if there's any problem and instead `None` will
            be returned as the value.
        :return: List with the ABI decoded return values
        """
        if not contract_functions:
            return []
        payloads = []
        params: TxParams = {'gas': Wei(0), 'gasPrice': Wei(0)}
        for _, contract_function in enumerate(contract_functions):
            if not contract_function.address:
                raise ValueError(f'Missing address for batch_call in `{contract_function.fn_name}`')

            payload = {'to': contract_function.address,
                       'data': contract_function.buildTransaction(params)['data'],
                       'output_type': [output['type'] for output in contract_function.abi['outputs']],
                       'fn_name': contract_function.fn_name,  # For debugging purposes
                       }
            if from_address:
                payload['from'] = from_address
            payloads.append(payload)

        return self.batch_call_custom(payloads, raise_exception=raise_exception, block_identifier=block_identifier)

    def estimate_gas(self, to: str, from_: Optional[str] = None, value: Optional[int] = None,
                     data: Optional[EthereumData] = None, gas: Optional[int] = None, gas_price: Optional[int] = None,
                     block_identifier: Optional[BlockIdentifier] = None) -> int:
        """
        Estimate gas calling `eth_estimateGas`

        :param from_:
        :param to:
        :param value:
        :param data:
        :param gas:
        :param gas_price:
        :param block_identifier: Be careful, `Geth` does not support `pending` when estimating
        :return: Amount of gas needed for transaction
        :raises: ValueError
        """
        tx: TxParams = {'to': to}
        if from_:
            tx['from'] = from_
        if value:
            tx['value'] = value
        if data:
            tx['data'] = data
        if gas:
            tx['gas'] = gas
        if gas_price:
            tx['gasPrice'] = gas_price
        try:
            return self.w3.eth.estimate_gas(tx, block_identifier=block_identifier)
        except ValueError:
            if block_identifier is not None:  # Geth does not support setting `block_identifier`
                return self.w3.eth.estimate_gas(tx, block_identifier=None)
            else:
                raise

    @staticmethod
    def estimate_data_gas(data: bytes):
        if isinstance(data, str):
            data = HexBytes(data)

        gas = 0
        for byte in data:
            if not byte:
                gas += GAS_CALL_DATA_ZERO_BYTE
            else:
                gas += GAS_CALL_DATA_BYTE
        return gas

    def get_balance(self, address: ChecksumAddress, block_identifier: Optional[BlockIdentifier] = None):
        return self.w3.eth.get_balance(address, block_identifier)

    def get_transaction(self, tx_hash: EthereumHash) -> Optional[TxData]:
        try:
            return self.w3.eth.get_transaction(tx_hash)
        except TransactionNotFound:
            return None

    def get_transactions(self, tx_hashes: List[EthereumHash]) -> List[Optional[TxData]]:
        if not tx_hashes:
            return []
        payload = [{'id': i, 'jsonrpc': '2.0', 'method': 'eth_getTransactionByHash',
                    'params': [HexBytes(tx_hash).hex()]}
                   for i, tx_hash in enumerate(tx_hashes)]
        results = self.http_session.post(self.ethereum_node_url, json=payload).json()
        txs = []
        for result in results:
            raw_tx = result['result']
            if raw_tx:
                txs.append(transaction_formatter(raw_tx))
            else:
                txs.append(None)
        return txs

    def get_transaction_receipt(self, tx_hash: EthereumHash, timeout=None) -> Optional[TxReceipt]:
        try:
            if not timeout:
                tx_receipt = self.w3.eth.get_transaction_receipt(tx_hash)
            else:
                try:
                    tx_receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash, timeout=timeout)
                except TimeExhausted:
                    return None

            # Parity returns tx_receipt even is tx is still pending, so we check `blockNumber` is not None
            return tx_receipt if tx_receipt and tx_receipt['blockNumber'] is not None else None
        except TransactionNotFound:
            return None

    def get_transaction_receipts(self, tx_hashes: Sequence[EthereumData]) -> List[Optional[TxReceipt]]:
        if not tx_hashes:
            return []
        payload = [{'id': i, 'jsonrpc': '2.0', 'method': 'eth_getTransactionReceipt',
                    'params': [HexBytes(tx_hash).hex()]}
                   for i, tx_hash in enumerate(tx_hashes)]
        results = self.http_session.post(self.ethereum_node_url, json=payload).json()
        receipts = []
        for result in results:
            tx_receipt = result['result']
            # Parity returns tx_receipt even is tx is still pending, so we check `blockNumber` is not None
            if tx_receipt and tx_receipt['blockNumber'] is not None:
                receipts.append(receipt_formatter(tx_receipt))
            else:
                receipts.append(None)
        return receipts

    def get_block(self, block_identifier: BlockIdentifier, full_transactions: bool = False) -> Optional[BlockData]:
        try:
            return self.w3.eth.get_block(block_identifier, full_transactions=full_transactions)
        except BlockNotFound:
            return None

    def get_blocks(self, block_identifiers: Iterable[BlockIdentifier],
                   full_transactions: bool = False) -> List[Optional[BlockData]]:
        if not block_identifiers:
            return []
        payload = [{'id': i, 'jsonrpc': '2.0', 'method': 'eth_getBlockByNumber',
                    'params': [hex(block_identifier) if isinstance(block_identifier, int) else block_identifier,
                               full_transactions]}
                   for i, block_identifier in enumerate(block_identifiers)]
        results = self.http_session.post(self.ethereum_node_url, json=payload).json()
        blocks = []
        for result in results:
            raw_block = result['result']
            if raw_block:
                if 'extraData' in raw_block:
                    del raw_block['extraData']  # Remove extraData, raises some problems on parsing
                blocks.append(block_formatter(raw_block))
            else:
                blocks.append(None)
        return blocks

    def is_contract(self, contract_address: ChecksumAddress) -> bool:
        return bool(self.w3.eth.get_code(contract_address))

    @tx_with_exception_handling
    def send_transaction(self, transaction_dict: TxParams) -> HexBytes:
        return self.w3.eth.send_transaction(transaction_dict)

    @tx_with_exception_handling
    def send_raw_transaction(self, raw_transaction: EthereumData) -> HexBytes:
        return self.w3.eth.send_raw_transaction(bytes(raw_transaction))

    def send_unsigned_transaction(self, tx: TxParams, private_key: Optional[str] = None,
                                  public_key: Optional[str] = None, retry: bool = False,
                                  block_identifier: Optional[BlockIdentifier] = 'pending') -> HexBytes:
        """
        Send a tx using an unlocked public key in the node or a private key. Both `public_key` and
        `private_key` cannot be `None`

        :param tx:
        :param private_key:
        :param public_key:
        :param retry: Retry if a problem with nonce is found
        :param block_identifier: For nonce calculation, recommended is `pending`
        :return: tx hash
        """

        # TODO Refactor this method, it's not working well with new version of the nodes
        if private_key:
            address = self.private_key_to_address(private_key)
        elif public_key:
            address = public_key
        else:
            logger.error('No ethereum account provided. Need a public_key or private_key')
            raise ValueError('Ethereum account was not configured or unlocked in the node')

        if tx.get('nonce') is None:
            tx['nonce'] = self.get_nonce_for_account(address, block_identifier=block_identifier)

        number_errors = 5
        while number_errors >= 0:
            try:
                if private_key:
                    signed_tx = self.w3.eth.account.sign_transaction(tx, private_key=private_key)
                    logger.debug('Sending %d wei from %s to %s', tx['value'], address, tx['to'])
                    try:
                        return self.send_raw_transaction(signed_tx.rawTransaction)
                    except TransactionAlreadyImported as e:
                        # Sometimes Parity 2.2.11 fails with Transaction already imported, even if it's not, but it's
                        # processed
                        tx_hash = signed_tx.hash
                        logger.error('Transaction with tx-hash=%s already imported: %s' % (tx_hash.hex(), str(e)))
                        return tx_hash
                elif public_key:
                    tx['from'] = address
                    return self.send_transaction(tx)
            except ReplacementTransactionUnderpriced as e:
                if not retry or not number_errors:
                    raise e
                current_nonce = tx['nonce']
                tx['nonce'] = max(current_nonce + 1, self.get_nonce_for_account(address,
                                                                                block_identifier=block_identifier))
                logger.error('Tx with nonce=%d was already sent for address=%s, retrying with nonce=%s',
                             current_nonce, address, tx['nonce'])
            except InvalidNonce as e:
                if not retry or not number_errors:
                    raise e
                logger.error('address=%s Tx with invalid nonce=%d, retrying recovering nonce again',
                             address, tx['nonce'])
                tx['nonce'] = self.get_nonce_for_account(address, block_identifier=block_identifier)
                number_errors -= 1

    def send_eth_to(self, private_key: str, to: str, gas_price: int, value: Wei, gas: int = 23000,
                    nonce: Optional[int] = None, retry: bool = False,
                    block_identifier: Optional[BlockIdentifier] = 'pending') -> bytes:
        """
        Send ether using configured account

        :param to: to
        :param gas_price: gas_price
        :param value: value(wei)
        :param gas: gas, defaults to 22000
        :param retry: Retry if a problem is found
        :param nonce: Nonce of sender account
        :param block_identifier: Block identifier for nonce calculation
        :return: tx_hash
        """

        assert check_checksum(to)

        tx: TxParams = {
            'to': to,
            'value': value,
            'gas': Wei(gas),
            'gasPrice': Wei(gas_price),
        }

        if nonce is not None:
            tx['nonce'] = Nonce(nonce)

        return self.send_unsigned_transaction(tx, private_key=private_key, retry=retry,
                                              block_identifier=block_identifier)

    def check_tx_with_confirmations(self, tx_hash: EthereumHash, confirmations: int) -> bool:
        """
        Check tx hash and make sure it has the confirmations required

        :param tx_hash: Hash of the tx
        :param confirmations: Minimum number of confirmations required
        :return: True if tx was mined with the number of confirmations required, False otherwise
        """
        tx_receipt = self.get_transaction_receipt(tx_hash)
        if not tx_receipt or tx_receipt['blockNumber'] is None:
            # If `tx_receipt` exists but `blockNumber` is `None`, tx is still pending (just Parity)
            return False
        else:
            return (self.w3.eth.block_number - tx_receipt['blockNumber']) >= confirmations

    @staticmethod
    def private_key_to_address(private_key):
        return checksum_encode(privtoaddr(private_key))



import logging

from django.conf import settings

from eth_account import Account
from eth_account.signers.local import LocalAccount
from web3 import Web3
from web3.contract import Contract

from ..ethereum_client import EthereumClient, EthereumClientProvider
from .utils import deploy_erc20, deploy_example_erc20, send_tx

logger = logging.getLogger(__name__)


class EthereumTestCaseMixin:
    ethereum_client: EthereumClient
    w3: Web3
    ethereum_test_account: LocalAccount

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.ethereum_client = EthereumClientProvider()
        cls.w3 = cls.ethereum_client.w3
        cls.ethereum_test_account = Account.from_key(settings.ETHEREUM_TEST_PRIVATE_KEY)

    @property
    def gas_price(self):
        return self.w3.eth.gas_price

    def send_tx(self, tx, account: LocalAccount) -> bytes:
        return send_tx(self.w3, tx, account)

    def send_ether(self, to: str, value: int) -> bytes:
        return send_tx(self.w3, {'to': to, 'value': value}, self.ethereum_test_account)

    def create_account(self, initial_ether: float = 0, initial_wei: int = 0) -> LocalAccount:
        account = Account.create()
        if initial_ether > .0 or initial_wei > 0:
            self.send_tx({'to': account.address,
                          'value': self.w3.toWei(initial_ether, 'ether') + initial_wei
                          }, self.ethereum_test_account)
        return account

    def deploy_erc20(self, name: str, symbol: str, owner: str, amount: int, decimals: int = 18, deployer: str = None,
                     account: LocalAccount = None) -> Contract:
        return deploy_erc20(self.w3, name, symbol, owner, amount, decimals=decimals, deployer=deployer,
                            account=account)

    def deploy_example_erc20(self, amount: int, owner: str) -> Contract:
        return deploy_example_erc20(self.w3, amount, owner, account=self.ethereum_test_account)
